# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: MIT

from collections.abc import Hashable
import copy
from typing import Any, Iterable, Optional, Tuple, Union


# -------------------------------------------------------------------------------------
ListKeyVal = Union[str, Tuple[str, ...]]


class KeyedList(list):
    """
    YANG lists are not ordered by default but indexed by keys. This class mimics some
    parts of the list API but allows direct access to elements when their keys are
    known.

    Note that even though this class inherits from list, it is not really a list.
    Internally, it behaves as a dict. Some parts of the list API that assume any
    ordering are not supported for that reason.

    The inheritance is only to avoid confusion and make isinstance return something that
    looks like a list.
    """

    def __init__(
        self,
        init: Iterable = None,
        key_name: Optional[Union[str, Tuple[str, ...]]] = None,
    ) -> None:
        """
        :arg init:
            Values to initialize the list with.
        :arg key_name:
            Name of keys in the element when the element is a dict. If there is only one
            key, it is a plain string, if there is more than one key, it is a tuple of
            strings. If there is no key (leaf-list), key_name is None.
        """
        super().__init__()  # this is not a real list anyway
        self._key_name = key_name
        self._map = {}  # type: Dict[ListKeyVal, Any]
        if init is not None:
            self.extend(init)

    def _element_key(self, element: Any) -> Hashable:
        if self._key_name is None:
            return py_to_yang(element)
        if not isinstance(element, dict):
            raise TypeError("element must be a dict")
        if isinstance(self._key_name, str):
            return py_to_yang(element[self._key_name])
        return tuple(py_to_yang(element[k]) for k in self._key_name)

    def clear(self) -> None:
        self._map.clear()

    def copy(self) -> "KeyedList":
        return KeyedList(self._map.values(), key_name=self._key_name)

    def append(self, element: Any) -> None:
        key = self._element_key(element)
        if key in self._map:
            raise ValueError("element with key %r already in list" % (key,))
        self._map[key] = element

    def extend(self, iterable: Iterable) -> None:
        for element in iterable:
            self.append(element)

    def pop(self, key: ListKeyVal = None, default: Any = None) -> Any:
        if key is None or isinstance(key, (int, slice)):
            raise TypeError("non-ordered lists cannot be accessed by index")
        return self._map.pop(key, default)

    def remove(self, element: Any) -> None:
        key = self._element_key(element)
        del self._map[key]

    def __getitem__(self, key: ListKeyVal) -> Any:
        if isinstance(key, (int, slice)):
            raise TypeError("non-ordered lists cannot be accessed by index")
        return self._map[key]

    def __delitem__(self, key: ListKeyVal) -> None:
        if isinstance(key, (int, slice)):
            raise TypeError("non-ordered lists cannot be accessed by index")
        del self._map[key]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, KeyedList):
            return other._map == self._map
        if not isinstance(other, list):
            return False
        try:
            other_map = {}
            for e in other:
                other_map[self._element_key(e)] = e
            return other_map == self._map
        except (KeyError, TypeError):
            return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __iter__(self):
        return iter(self._map.values())

    def __len__(self):
        return len(self._map)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return str(list(self._map.values()))

    def count(self, element: Any):
        if element in self:
            return 1
        return 0

    def __contains__(self, element: Any) -> bool:
        try:
            if isinstance(element, dict) or element not in self._map:
                key = self._element_key(element)
            else:
                key = element
            return key in self._map
        except (KeyError, TypeError):
            return False

    def __copy__(self) -> "KeyedList":
        return self.copy()

    def __deepcopy__(self, memo) -> "KeyedList":
        k = KeyedList.__new__(KeyedList)
        memo[id(self)] = k
        k._key_name = copy.deepcopy(self._key_name, memo)
        k._map = copy.deepcopy(self._map, memo)
        return k

    # unsupported list API methods
    def __unsupported(self, *args, **kwargs):
        raise TypeError("unsupported operation for non-ordered lists")

    index = __unsupported
    insert = __unsupported
    reverse = __unsupported
    sort = __unsupported
    __add__ = __unsupported
    __ge__ = __unsupported
    __gt__ = __unsupported
    __iadd__ = __unsupported
    __imul__ = __unsupported
    __le__ = __unsupported
    __lt__ = __unsupported
    __mul__ = __unsupported
    __reversed__ = __unsupported
    __rmul__ = __unsupported
    __setitem__ = __unsupported


# -------------------------------------------------------------------------------------
def py_to_yang(val: Any) -> str:
    """
    Convert a python value to a string following how it would be stored in a libyang
    data tree. Also suitable for comparison with YANG list keys (which are always
    strings).
    """
    if isinstance(val, str):
        return val
    if val is True:
        return "true"
    if val is False:
        return "false"
    return str(val)



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: MIT

import copy
import unittest

import libyang as ly


# -------------------------------------------------------------------------------------
class XPathTest(unittest.TestCase):
    def test_xpath_split(self):
        for xpath, split_result in XPATH_SPLIT_EXPECTED_RESULTS.items():
            if split_result is ValueError:
                with self.assertRaises(split_result):
                    list(ly.xpath_split(xpath))
            else:
                self.assertEqual(list(ly.xpath_split(xpath)), split_result)

    def test_xpath_get(self):
        for xpath, value, defval, expected in XPATH_GET_EXPECTED_RESULTS:
            res = ly.xpath_get(DICT, xpath, defval)
            if expected:
                self.assertEqual(res, value)
                self.assertNotEqual(res, defval)
            else:
                self.assertNotEqual(res, value)
                self.assertEqual(res, defval)

    def test_xpath_set(self):
        d = copy.deepcopy(DICT)
        ly.xpath_set(d, "/val", 43)
        ly.xpath_set(d, "/cont1/leaf1", "foo")
        ly.xpath_set(d, "/cont1/leaf2", "bar")
        ly.xpath_set(d, "/iface[name='eth0']", {"name": "eth0", "up": False})
        ly.xpath_set(d, "/iface[name='eth1']/ipv4/address[ip='10.0.0.2']/mtu", 1500)
        ly.xpath_set(
            d,
            "/iface[name='eth1']/ipv6/address[ip='3ffe::ffff'][prefixlen='100']",
            {"ip": "3ffe::ffff", "prefixlen": 100, "tentative": False},
        )
        ly.xpath_set(d, "/lstnum[.='100']", 100)
        ly.xpath_set(d, "/lstnum[.='1']", 1, after="")
        ly.xpath_set(d, "/lstnum[5]", 33, after="4")
        ly.xpath_set(d, "/lstnum[5]", 34, after="4")
        ly.xpath_set(d, "/lstnum[5]", 35, after="4")
        ly.xpath_set(d, "/lstnum[7]", 101, after="6")
        ly.xpath_set(d, "/lstnum[8]", 102, after="7")
        with self.assertRaises(ValueError):
            ly.xpath_set(d, "/lstnum[.='1000']", 1000, after="1000000")
        with self.assertRaises(ValueError):
            ly.xpath_set(
                d,
                "/iface[name='eth1']/ipv4/address[ip='10.0.0.3']",
                {"ip": "10.0.0.3"},
                after="[ip='10.0.0.55']",
            )
        ly.xpath_set(
            d, "/iface[name='eth1']/ipv4/address[ip='10.0.0.3']", {"ip": "10.0.0.3"}
        )
        ly.xpath_set(d, "/cont2/newlist[foo='bar']", {"foo": "bar", "name": "baz"})
        ly.xpath_set(d, "/cont2/newll[.='12']", 12)
        self.assertEqual(
            d,
            {
                "cont1": {"leaf1": "foo", "leaf2": "bar"},
                "cont2": {
                    "leaf2": "coucou2",
                    "newlist": [{"foo": "bar", "name": "baz"}],
                    "newll": [12],
                },
                "iface": [
                    {"name": "eth0", "up": False},
                    {
                        "name": "eth1",
                        "ipv4": {
                            "address": [
                                {"ip": "10.0.0.2", "mtu": 1500},
                                {"ip": "10.0.0.6"},
                                {"ip": "10.0.0.3"},
                            ]
                        },
                        "ipv6": {
                            "address": [
                                {
                                    "ip": "3ffe::321:8",
                                    "prefixlen": 64,
                                    "tentative": False,
                                },
                                {
                                    "ip": "3ffe::ff12",
                                    "prefixlen": 96,
                                    "tentative": True,
                                },
                                {
                                    "ip": "3ffe::ffff",
                                    "prefixlen": 100,
                                    "tentative": False,
                                },
                            ]
                        },
                    },
                ],
                "iface2": [
                    {"name": "eth2", "mtu": 1500},
                    {"name": "eth3", "mtu": 1000},
                ],
                "lst2": ["a", "b", "c"],
                "lstnum": [1, 10, 20, 30, 35, 100, 101, 102],
                "val": 43,
            },
        )

    def test_xpath_del(self):
        d = copy.deepcopy(DICT)
        self.assertTrue(ly.xpath_del(d, "/val"))
        self.assertTrue(ly.xpath_del(d, "/cont2/leaf2"))
        self.assertTrue(ly.xpath_del(d, "/lst2[.='b']"))
        self.assertTrue(ly.xpath_del(d, "/lst2[.='a']"))
        self.assertTrue(ly.xpath_del(d, "/lst2[.='c']"))
        self.assertFalse(ly.xpath_del(d, "/lst2[.='x']"))
        self.assertFalse(ly.xpath_del(d, "/lst2"))
        self.assertTrue(ly.xpath_del(d, "lstnum[.='20']"))
        self.assertFalse(ly.xpath_del(d, "/lstnum[.='10000']"))
        self.assertFalse(ly.xpath_del(d, "/foo/bar/baz"))
        self.assertTrue(
            ly.xpath_del(d, "/iface[name='eth1']/ipv4/address[ip='10.0.0.2']")
        )
        self.assertFalse(
            ly.xpath_del(d, "/iface[name='eth1']/ipv4/address[ip='10.0.0.7']")
        )
        self.assertFalse(
            ly.xpath_del(
                d, "/iface[name='eth1']/ipv6/address[ip='3ffe::ff12'][prefixlen='64']"
            )
        )
        self.assertTrue(
            ly.xpath_del(
                d, "/iface[name='eth1']/ipv6/address[ip='3ffe::ff12'][prefixlen='96']"
            )
        )
        self.assertEqual(
            d,
            {
                "cont1": {"leaf1": "coucou1"},
                "cont2": {},
                "iface": [
                    {
                        "name": "eth0",
                        "ipv4": {"address": [{"ip": "10.0.0.1"}, {"ip": "10.0.0.153"}]},
                        "ipv6": {
                            "address": [{"ip": "3ffe::123:1"}, {"ip": "3ffe::c00:c00"}]
                        },
                    },
                    {
                        "name": "eth1",
                        "ipv4": {"address": [{"ip": "10.0.0.6"}]},
                        "ipv6": {
                            "address": [
                                {
                                    "ip": "3ffe::321:8",
                                    "prefixlen": 64,
                                    "tentative": False,
                                }
                            ]
                        },
                    },
                ],
                "iface2": [
                    {"name": "eth2", "mtu": 1500},
                    {"name": "eth3", "mtu": 1000},
                ],
                "lstnum": [10, 30, 40],
            },
        )

    def test_xpath_move(self):
        d = {
            "regular": [
                {"name": "foo", "size": 2},
                {"name": "bar", "size": 1},
                {"name": "baz", "size": 42},
            ],
            "regular-multi": [
                {"name": "foo", "group": "xxx", "size": 2},
                {"name": "bar", "group": "yyy", "size": 1},
                {"name": "baz", "group": "xxx", "size": 42},
            ],
            "ll": ["a", "b", "c", "d"],
            "ll-keyed": ly.KeyedList([10, 20, 30, 40]),
        }
        with self.assertRaises(KeyError):
            ly.xpath_move(d, "/not/found", "")
        with self.assertRaises(KeyError):
            ly.xpath_move(d, "/ll[.='x']", "")
        with self.assertRaises(KeyError):
            ly.xpath_move(d, "/ll[.='c']", "x")
        with self.assertRaises(ValueError):
            ly.xpath_move(d, "/ll", "")
        ly.xpath_move(d, "/ll[.='c']", "a")
        ly.xpath_move(d, "/ll[.='b']", "")
        ly.xpath_move(d, "/ll[.='a']", None)
        with self.assertRaises(ValueError):
            ly.xpath_move(d, "/ll-keyed[.='40']", "")
        with self.assertRaises(ValueError):
            ly.xpath_move(d, "/ll-keyed[.='10']", None)
        with self.assertRaises(ValueError):
            ly.xpath_move(d, "/ll-keyed[.='30']", "40")
        self.assertEqual(
            d,
            {
                "regular": [
                    {"name": "foo", "size": 2},
                    {"name": "bar", "size": 1},
                    {"name": "baz", "size": 42},
                ],
                "regular-multi": [
                    {"group": "xxx", "name": "foo", "size": 2},
                    {"group": "yyy", "name": "bar", "size": 1},
                    {"group": "xxx", "name": "baz", "size": 42},
                ],
                "ll": ["b", "c", "d", "a"],
                "ll-keyed": [10, 20, 30, 40],
            },
        )

    def test_xpath_getall(self):
        for xpath, expected in XPATH_GETALL_EXPECTED_RESULTS:
            # result order is not guaranteed, and we cannot use a set() because
            # dicts are not hashable.
            res = list(ly.xpath_getall(DICT, xpath))
            self.assertEqual(len(res), len(expected), xpath)
            for elt in res:
                self.assertIn(elt, expected, xpath)


XPATH_SPLIT_EXPECTED_RESULTS = {
    "": ValueError,
    "/p:nam": [("p", "nam", [])],
    "/nam1/nam2": [(None, "nam1", []), (None, "nam2", [])],
    "/nam1/p:nam2": [(None, "nam1", []), ("p", "nam2", [])],
    '/p:nam/lst[k1="foo"]': [("p", "nam", []), (None, "lst", [("k1", "foo")])],
    '/p:nam/lst[.="foo"]': [("p", "nam", []), (None, "lst", [(".", "foo")])],
    '/p:nam/lst[.="foo\\bar"]': [("p", "nam", []), (None, "lst", [(".", "foo\\bar")])],
    "/nam/p:lst[k1='foo'][k2='bar']/x": [
        (None, "nam", []),
        ("p", "lst", [("k1", "foo"), ("k2", "bar")]),
        (None, "x", []),
    ],
    "/nam/p:lst[k1='=:[/]\\'']/l2[k2=\"dead::beef/64\"]": [
        (None, "nam", []),
        ("p", "lst", [("k1", "=:[/]'")]),
        (None, "l2", [("k2", "dead::beef/64")]),
    ],
    "foo": [(None, "foo", [])],
    "foo/bar": [(None, "foo", []), (None, "bar", [])],
    "//invalid/xpath": ValueError,
    "/xxx[abc='2']invalid:xx/xpath": ValueError,
}

DICT = {
    "cont1": {
        "leaf1": "coucou1",
    },
    "cont2": {
        "leaf2": "coucou2",
    },
    "iface": [
        {
            "name": "eth0",
            "ipv4": {
                "address": [
                    {"ip": "10.0.0.1"},
                    {"ip": "10.0.0.153"},
                ],
            },
            "ipv6": {
                "address": [
                    {"ip": "3ffe::123:1"},
                    {"ip": "3ffe::c00:c00"},
                ],
            },
        },
        {
            "name": "eth1",
            "ipv4": {
                "address": ly.KeyedList(
                    [
                        {"ip": "10.0.0.2"},
                        {"ip": "10.0.0.6"},
                    ],
                    key_name="ip",
                ),
            },
            "ipv6": {
                "address": ly.KeyedList(
                    [
                        {"ip": "3ffe::321:8", "prefixlen": 64, "tentative": False},
                        {"ip": "3ffe::ff12", "prefixlen": 96, "tentative": True},
                    ],
                    key_name=("ip", "prefixlen"),
                ),
            },
        },
    ],
    "iface2": ly.KeyedList(
        [{"name": "eth2", "mtu": 1500}, {"name": "eth3", "mtu": 1000}], key_name="name"
    ),
    "lst2": ["a", "b", "c"],
    "lstnum": [10, 20, 30, 40],
    "val": 42,
}

XPATH_GET_EXPECTED_RESULTS = [
    ("/val", 42, None, True),
    ("val", 42, None, True),
    ("lst2", ["a", "b", "c"], None, True),
    (
        "iface[name='eth0']/ipv4/address",
        [{"ip": "10.0.0.1"}, {"ip": "10.0.0.153"}],
        None,
        True,
    ),
    (
        "/iface[name='eth1']/ipv6/address[ip='3ffe::321:8'][prefixlen='64']",
        {"ip": "3ffe::321:8", "prefixlen": 64, "tentative": False},
        None,
        True,
    ),
    ("cont1/leaf1", "coucou1", None, True),
    ("cont2/leaf2", "coucou2", None, True),
    ("cont1/leaf2", "not found", "fallback", False),
    ("cont1/leaf2", "not found", None, False),
]

XPATH_GETALL_EXPECTED_RESULTS = [
    ("/val", [42]),
    ("val", [42]),
    ("lst2", ["a", "b", "c"]),
    ("/lstnum", [10, 20, 30, 40]),
    ("iface/name", ["eth0", "eth1"]),
    ("/iface*/name", ["eth0", "eth1", "eth2", "eth3"]),
    ('iface[name="eth1"]/ipv4/address/ip', ["10.0.0.2", "10.0.0.6"]),
    (
        "iface/ipv*/address/ip",
        [
            "10.0.0.1",
            "10.0.0.153",
            "3ffe::123:1",
            "3ffe::c00:c00",
            "10.0.0.2",
            "10.0.0.6",
            "3ffe::321:8",
            "3ffe::ff12",
        ],
    ),
    (
        "/iface[name='eth1']/ipv6/address[ip='3ffe::321:8'][prefixlen='64']/tentative",
        [False],
    ),
    (
        "/iface[name='eth1']/ipv6/address[ip='3ffe::8'][prefixlen='64']/tentative",
        [],
    ),
    (
        "/cont1[foo='bar']",
        [],
    ),
    (
        "/iface*",
        [
            {
                "name": "eth0",
                "ipv4": {
                    "address": [
                        {"ip": "10.0.0.1"},
                        {"ip": "10.0.0.153"},
                    ],
                },
                "ipv6": {
                    "address": [
                        {"ip": "3ffe::123:1"},
                        {"ip": "3ffe::c00:c00"},
                    ],
                },
            },
            {
                "name": "eth1",
                "ipv4": {
                    "address": [
                        {"ip": "10.0.0.2"},
                        {"ip": "10.0.0.6"},
                    ],
                },
                "ipv6": {
                    "address": [
                        {"ip": "3ffe::321:8", "prefixlen": 64, "tentative": False},
                        {"ip": "3ffe::ff12", "prefixlen": 96, "tentative": True},
                    ],
                },
            },
            {"name": "eth2", "mtu": 1500},
            {"name": "eth3", "mtu": 1000},
        ],
    ),
    (
        "/iface*[name='eth0']/ipv4",
        [
            {
                "address": [
                    {"ip": "10.0.0.1"},
                    {"ip": "10.0.0.153"},
                ],
            },
        ],
    ),
    (
        "/iface[name='eth0']/ipv4/address[ip='10.0.0.1']",
        [{"ip": "10.0.0.1"}],
    ),
    (
        "/iface[name='eth0']/ipv4/address[ip='10.0.0.2']",
        [],
    ),
    ("/iface2[name='eth3']", [{"name": "eth3", "mtu": 1000}]),
    ("/iface[name='eth0']/ipv4/address", [{"ip": "10.0.0.1"}, {"ip": "10.0.0.153"}]),
    ("*/leaf*", ["coucou1", "coucou2"]),
    ("cont1/not-found", []),
    (
        "iface/ipv4",
        [
            {"address": [{"ip": "10.0.0.1"}, {"ip": "10.0.0.153"}]},
            {"address": [{"ip": "10.0.0.2"}, {"ip": "10.0.0.6"}]},
        ],
    ),
    ('iface*[name="eth3"]', [{"name": "eth3", "mtu": 1000}]),
    (
        'iface*[name="eth0"]',
        [
            {
                "name": "eth0",
                "ipv4": {
                    "address": [
                        {"ip": "10.0.0.1"},
                        {"ip": "10.0.0.153"},
                    ],
                },
                "ipv6": {
                    "address": [
                        {"ip": "3ffe::123:1"},
                        {"ip": "3ffe::c00:c00"},
                    ],
                },
            }
        ],
    ),
]



# Copyright (c) 2018-2019 Robin Jarry
# Copyright (c) 2020 6WIND S.A.
# Copyright (c) 2021 RACOM s.r.o.
# SPDX-License-Identifier: MIT


from .context import Context
from .data import (
    DAnydata,
    DAnyxml,
    DContainer,
    DLeaf,
    DLeafList,
    DList,
    DNode,
    DNodeAttrs,
    DNotif,
    DRpc,
)
from .diff import (
    BaseTypeAdded,
    BaseTypeRemoved,
    BitAdded,
    BitRemoved,
    BitStatusAdded,
    BitStatusRemoved,
    ConfigFalseAdded,
    ConfigFalseRemoved,
    DefaultAdded,
    DefaultRemoved,
    DescriptionAdded,
    DescriptionRemoved,
    EnumAdded,
    EnumRemoved,
    EnumStatusAdded,
    EnumStatusRemoved,
    ExtensionAdded,
    ExtensionRemoved,
    KeyAdded,
    KeyRemoved,
    LengthAdded,
    LengthRemoved,
    MandatoryAdded,
    MandatoryRemoved,
    MustAdded,
    MustRemoved,
    NodeTypeAdded,
    NodeTypeRemoved,
    OrderedByUserAdded,
    OrderedByUserRemoved,
    PatternAdded,
    PatternRemoved,
    PresenceAdded,
    PresenceRemoved,
    RangeAdded,
    RangeRemoved,
    SNodeAdded,
    SNodeDiff,
    SNodeRemoved,
    StatusAdded,
    StatusRemoved,
    UnitsAdded,
    UnitsRemoved,
    schema_diff,
)
from .extension import ExtensionPlugin, LibyangExtensionError
from .keyed_list import KeyedList
from .log import configure_logging, temp_log_options
from .schema import (
    Enum,
    Extension,
    ExtensionCompiled,
    ExtensionParsed,
    Feature,
    Identity,
    IfAndFeatures,
    IfFeature,
    IfFeatureExpr,
    IfFeatureExprTree,
    IfNotFeature,
    IfOrFeatures,
    Module,
    Must,
    PAction,
    PActionInOut,
    PAnydata,
    Pattern,
    PAugment,
    PCase,
    PChoice,
    PContainer,
    PEnum,
    PGrouping,
    PIdentity,
    PLeaf,
    PLeafList,
    PList,
    PNode,
    PNotif,
    PRefine,
    PType,
    PUses,
    Revision,
    SAnydata,
    SAnyxml,
    SCase,
    SChoice,
    SContainer,
    SLeaf,
    SLeafList,
    SList,
    SNode,
    SRpc,
    SRpcInOut,
    Type,
    Typedef,
)
from .util import DataType, IOType, LibyangError
from .xpath import (
    xpath_del,
    xpath_get,
    xpath_getall,
    xpath_move,
    xpath_set,
    xpath_setdefault,
    xpath_split,
)


__all__ = (
    "BaseTypeAdded",
    "BaseTypeRemoved",
    "BitAdded",
    "BitRemoved",
    "ConfigFalseAdded",
    "ConfigFalseRemoved",
    "Context",
    "DContainer",
    "DLeaf",
    "DLeafList",
    "DList",
    "DNode",
    "DRpc",
    "DefaultAdded",
    "DefaultRemoved",
    "DescriptionAdded",
    "DescriptionRemoved",
    "Enum",
    "EnumAdded",
    "EnumRemoved",
    "Extension",
    "ExtensionAdded",
    "ExtensionCompiled",
    "ExtensionParsed",
    "ExtensionPlugin",
    "ExtensionRemoved",
    "Feature",
    "Identity",
    "IfAndFeatures",
    "IfFeature",
    "IfFeatureExpr",
    "IfFeatureExprTree",
    "IfNotFeature",
    "IfOrFeatures",
    "KeyAdded",
    "KeyedList",
    "KeyRemoved",
    "LengthAdded",
    "LengthRemoved",
    "LibyangError",
    "IOType",
    "DataType",
    "MandatoryAdded",
    "MandatoryRemoved",
    "Module",
    "Must",
    "MustAdded",
    "MustRemoved",
    "NodeTypeAdded",
    "NodeTypeRemoved",
    "OrderedByUserAdded",
    "OrderedByUserRemoved",
    "PAction",
    "PActionInOut",
    "PAnydata",
    "PAugment",
    "PCase",
    "PChoice",
    "PContainer",
    "PEnum",
    "PGrouping",
    "PIdentity",
    "PLeaf",
    "PLeafList",
    "PList",
    "PNode",
    "PNotif",
    "PRefine",
    "PType",
    "PUses",
    "Pattern",
    "PatternAdded",
    "PatternRemoved",
    "PresenceAdded",
    "PresenceRemoved",
    "RangeAdded",
    "RangeRemoved",
    "Revision",
    "SAnydata",
    "SAnyxml",
    "SCase",
    "SChoice",
    "SContainer",
    "SLeaf",
    "SLeafList",
    "SList",
    "SNode",
    "SNodeAdded",
    "SNodeDiff",
    "SNodeRemoved",
    "SRpc",
    "SRpcInOut",
    "StatusAdded",
    "StatusRemoved",
    "Type",
    "UnitsAdded",
    "UnitsRemoved",
    "configure_logging",
    "schema_diff",
    "xpath_del",
    "xpath_get",
    "xpath_getall",
    "xpath_move",
    "xpath_set",
    "xpath_setdefault",
    "xpath_split",
)



# Copyright (c) 2018-2019 Robin Jarry
# Copyright (c) 2021 RACOM s.r.o.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
import enum
from typing import Iterable, Optional
import warnings

from _libyang import ffi, lib


# -------------------------------------------------------------------------------------
@dataclass(frozen=True)
class LibyangErrorItem:
    msg: Optional[str]
    data_path: Optional[str]
    schema_path: Optional[str]
    line: Optional[int]


# -------------------------------------------------------------------------------------
class LibyangError(Exception):
    def __init__(
        self, message: str, *args, errors: Optional[Iterable[LibyangErrorItem]] = None
    ):
        super().__init__(message, *args)
        self.message = message
        self.errors = tuple(errors or ())

    def __str__(self):
        return self.message


# -------------------------------------------------------------------------------------
def deprecated(old: str, new: str, removed_in: str) -> None:
    msg = "%s has been replaced by %s, it will be removed in version %s"
    msg %= (old, new, removed_in)
    warnings.warn(msg, DeprecationWarning, stacklevel=3)


# -------------------------------------------------------------------------------------
def str2c(s: Optional[str], encode: bool = True):
    if s is None:
        return ffi.NULL
    if encode and hasattr(s, "encode"):
        s = s.encode("utf-8")
    return ffi.new("char []", s)


# -------------------------------------------------------------------------------------
def c2str(c, decode: bool = True):
    if c == ffi.NULL:  # C type: "char *"
        return None
    s = ffi.string(c)
    if decode and hasattr(s, "decode"):
        s = s.decode("utf-8")
    return s


# -------------------------------------------------------------------------------------
def p_str2c(s: Optional[str], encode: bool = True):
    s_p = str2c(s, encode)
    return ffi.new("char **", s_p)


# -------------------------------------------------------------------------------------
def ly_array_count(cdata):
    if cdata == ffi.NULL:
        return 0
    return ffi.cast("uint64_t *", cdata)[-1]


# -------------------------------------------------------------------------------------
def ly_array_iter(cdata):
    for i in range(ly_array_count(cdata)):
        yield cdata[i]


# -------------------------------------------------------------------------------------
def ly_list_iter(cdata):
    item = cdata
    while item != ffi.NULL:
        yield item
        item = item.next


# -------------------------------------------------------------------------------------
class IOType(enum.Enum):
    FD = enum.auto()
    FILE = enum.auto()
    FILEPATH = enum.auto()
    MEMORY = enum.auto()


# -------------------------------------------------------------------------------------
class DataType(enum.Enum):
    DATA_YANG = lib.LYD_TYPE_DATA_YANG
    RPC_YANG = lib.LYD_TYPE_RPC_YANG
    NOTIF_YANG = lib.LYD_TYPE_NOTIF_YANG
    REPLY_YANG = lib.LYD_TYPE_REPLY_YANG
    RPC_NETCONF = lib.LYD_TYPE_RPC_NETCONF
    NOTIF_NETCONF = lib.LYD_TYPE_NOTIF_NETCONF
    REPLY_NETCONF = lib.LYD_TYPE_REPLY_NETCONF
    RPC_RESTCONF = lib.LYD_TYPE_RPC_RESTCONF
    NOTIF_RESTCONF = lib.LYD_TYPE_NOTIF_RESTCONF
    REPLY_RESTCONF = lib.LYD_TYPE_REPLY_RESTCONF


# -------------------------------------------------------------------------------------
def init_output(out_type, out_target, out_data):
    output = None
    if out_type == IOType.FD:
        ret = lib.ly_out_new_fd(out_target.fileno(), out_data)

    elif out_type == IOType.FILE:
        ret = lib.ly_out_new_file(out_target, out_data)

    elif out_type == IOType.FILEPATH:
        out_target = str2c(out_target)
        ret = lib.ly_out_new_filepath(out_target, out_data)

    elif out_type == IOType.MEMORY:
        output = ffi.new("char **")
        ret = lib.ly_out_new_memory(output, 0, out_data)

    else:
        raise ValueError("invalid output")

    return ret, output


# -------------------------------------------------------------------------------------
def data_load(in_type, in_data, data, data_keepalive, encode=True):
    if in_type == IOType.FD:
        ret = lib.ly_in_new_fd(in_data.fileno(), data)
    elif in_type == IOType.FILE:
        ret = lib.ly_in_new_file(in_data, data)
    elif in_type == IOType.FILEPATH:
        ret = lib.ly_in_new_filepath(str2c(in_data), len(in_data), data)
    elif in_type == IOType.MEMORY:
        c_str = str2c(in_data, encode=encode)
        data_keepalive.append(c_str)
        ret = lib.ly_in_new_memory(c_str, data)
    else:
        raise ValueError("invalid input")
    return ret



# Copyright (c) 2019 Robin Jarry
# SPDX-License-Identifier: MIT

import os
import unittest

from libyang import (
    BaseTypeAdded,
    BaseTypeRemoved,
    Context,
    DefaultAdded,
    EnumRemoved,
    EnumStatusAdded,
    EnumStatusRemoved,
    ExtensionAdded,
    NodeTypeAdded,
    NodeTypeRemoved,
    SNodeAdded,
    SNodeRemoved,
    StatusAdded,
    StatusRemoved,
    schema_diff,
)


OLD_YANG_DIR = os.path.join(os.path.dirname(__file__), "yang-old")
NEW_YANG_DIR = os.path.join(os.path.dirname(__file__), "yang")


# -------------------------------------------------------------------------------------
class DiffTest(unittest.TestCase):
    expected_diffs = frozenset(
        (
            (BaseTypeAdded, "/yolo-system:conf/speed"),
            (BaseTypeAdded, "/yolo-system:state/speed"),
            (BaseTypeRemoved, "/yolo-system:conf/speed"),
            (BaseTypeRemoved, "/yolo-system:state/speed"),
            (DefaultAdded, "/yolo-system:conf/speed"),
            (DefaultAdded, "/yolo-system:state/speed"),
            (NodeTypeAdded, "/yolo-system:conf/number"),
            (NodeTypeAdded, "/yolo-system:state/number"),
            (NodeTypeRemoved, "/yolo-system:conf/number"),
            (NodeTypeRemoved, "/yolo-system:state/number"),
            (SNodeAdded, "/yolo-system:conf/full"),
            (SNodeAdded, "/yolo-system:state/full"),
            (SNodeAdded, "/yolo-system:conf/hostname-ref"),
            (SNodeAdded, "/yolo-system:conf/url/enabled"),
            (SNodeAdded, "/yolo-system:conf/url/fetch"),
            (
                SNodeAdded,
                "/yolo-system:conf/url/fetch/input/timeout",
            ),
            (
                SNodeAdded,
                "/yolo-system:conf/url/fetch/output/result",
            ),
            (SNodeAdded, "/yolo-system:state/hostname-ref"),
            (SNodeAdded, "/yolo-system:state/url/enabled"),
            (SNodeAdded, "/yolo-system:state/url/fetch"),
            (
                SNodeAdded,
                "/yolo-system:state/url/fetch/input/timeout",
            ),
            (
                SNodeAdded,
                "/yolo-system:state/url/fetch/output/result",
            ),
            (StatusAdded, "/yolo-system:conf/deprecated-leaf"),
            (StatusAdded, "/yolo-system:conf/obsolete-leaf"),
            (StatusAdded, "/yolo-system:state/deprecated-leaf"),
            (StatusAdded, "/yolo-system:state/obsolete-leaf"),
            (StatusRemoved, "/yolo-system:conf/deprecated-leaf"),
            (StatusRemoved, "/yolo-system:conf/obsolete-leaf"),
            (StatusRemoved, "/yolo-system:state/deprecated-leaf"),
            (StatusRemoved, "/yolo-system:state/obsolete-leaf"),
            (SNodeAdded, "/yolo-system:alarm-triggered"),
            (SNodeAdded, "/yolo-system:alarm-triggered/severity"),
            (SNodeAdded, "/yolo-system:alarm-triggered/description"),
            (SNodeAdded, "/yolo-system:config-change"),
            (SNodeAdded, "/yolo-system:config-change/edit"),
            (SNodeAdded, "/yolo-system:config-change/edit/target"),
            (EnumRemoved, "/yolo-system:conf/url/proto"),
            (EnumRemoved, "/yolo-system:state/url/proto"),
            (EnumStatusAdded, "/yolo-system:conf/url/proto"),
            (EnumStatusAdded, "/yolo-system:state/url/proto"),
            (ExtensionAdded, "/yolo-system:conf/url/proto"),
            (ExtensionAdded, "/yolo-system:state/url/proto"),
            (EnumStatusRemoved, "/yolo-system:conf/url/proto"),
            (EnumStatusRemoved, "/yolo-system:state/url/proto"),
            (SNodeAdded, "/yolo-system:conf/pill/red/out"),
            (SNodeAdded, "/yolo-system:state/pill/red/out"),
            (SNodeAdded, "/yolo-system:conf/pill/blue/in"),
            (SNodeAdded, "/yolo-system:state/pill/blue/in"),
            (SNodeAdded, "/yolo-system:alarm-triggered/severity"),
            (SNodeAdded, "/yolo-system:alarm-triggered/description"),
        )
    )

    def test_diff(self):
        with Context(OLD_YANG_DIR, compile_obsolete=True) as ctx_old, Context(
            NEW_YANG_DIR, compile_obsolete=True
        ) as ctx_new:
            mod = ctx_old.load_module("yolo-system")
            mod.feature_enable_all()
            mod = ctx_new.load_module("yolo-system")
            mod.feature_enable_all()
            diffs = []
            for d in schema_diff(ctx_old, ctx_new):
                if isinstance(d, (SNodeAdded, SNodeRemoved)):
                    diffs.append((d.__class__, d.node.schema_path()))
                else:
                    diffs.append((d.__class__, d.new.schema_path()))

            self.assertEqual(frozenset(diffs), self.expected_diffs)



# Copyright (c) 2018-2019 Robin Jarry
# Copyright (c) 2020 6WIND S.A.
# Copyright (c) 2021 RACOM s.r.o.
# SPDX-License-Identifier: MIT

import os
from typing import IO, Any, Callable, Iterator, List, Optional, Sequence, Tuple, Union

from _libyang import ffi, lib
from .data import (
    DNode,
    data_format,
    data_type,
    newval_flags,
    parser_flags,
    validation_flags,
)
from .schema import Module, SNode, schema_in_format
from .util import (
    DataType,
    IOType,
    LibyangError,
    LibyangErrorItem,
    c2str,
    data_load,
    str2c,
)


# -------------------------------------------------------------------------------------
@ffi.def_extern(name="lypy_module_imp_data_free_clb")
def libyang_c_module_imp_data_free_clb(cdata, user_data):
    instance = ffi.from_handle(user_data)
    instance.free_module_data(cdata)


# -------------------------------------------------------------------------------------
@ffi.def_extern(name="lypy_module_imp_clb")
def libyang_c_module_imp_clb(
    mod_name,
    mod_rev,
    submod_name,
    submod_rev,
    user_data,
    fmt,
    module_data,
    free_module_data,
):
    """
    Implement the C callback function for loading modules from any location.

    :arg c_str mod_name:
        The YANG module name
    :arg c_str mod_rev:
        The YANG module revision
    :arg c_str submod_name:
        The YANG submodule name
    :arg c_str submod_rev:
        The YANG submodule revision
    :arg user_data:
        The user data provided by user during registration. In this implementation
        it is always considered to be handle of Python object
    :arg fmt:
        The output pointer where to set the format of schema
    :arg module_data:
        The output pointer where to set the schema data itself
    :arg free_module_data:
        The output pointer of callback function which will be called when the schema
        data are no longer needed

    :returns:
        The LY_SUCCESS in case the needed YANG (sub)module schema was found
        The LY_ENOT in case the needed YANG (sub)module schema was not found
    """
    fmt[0] = lib.LYS_IN_UNKNOWN
    module_data[0] = ffi.NULL
    free_module_data[0] = lib.lypy_module_imp_data_free_clb
    instance = ffi.from_handle(user_data)
    ret = instance.get_module_data(
        c2str(mod_name), c2str(mod_rev), c2str(submod_name), c2str(submod_rev)
    )
    if ret is None:
        return lib.LY_ENOT
    in_fmt, content = ret
    fmt[0] = schema_in_format(in_fmt)
    module_data[0] = content
    return lib.LY_SUCCESS


# -------------------------------------------------------------------------------------
class ContextExternalModuleLoader:
    __slots__ = (
        "_cdata",
        "_module_data_clb",
        "_cffi_handle",
        "_cdata_modules",
    )

    def __init__(self, cdata) -> None:
        self._cdata = cdata  # C type: "struct ly_ctx *"
        self._module_data_clb = None
        self._cffi_handle = ffi.new_handle(self)
        self._cdata_modules = []

    def free_module_data(self, cdata) -> None:
        """
        Free previously stored data, obtained after a get_module_data.

        :arg cdata:
            The pointer to YANG modelu schema (c_str), which shall be released from memory
        """
        self._cdata_modules.remove(cdata)

    def get_module_data(
        self,
        mod_name: Optional[str],
        mod_rev: Optional[str],
        submod_name: Optional[str],
        submod_rev: Optional[str],
    ) -> Optional[Tuple[str, str]]:
        """
        Get the YANG module schema data based requirements from libyang_c_module_imp_clb
        function and forward that request to user Python based callback function.

        The returned data from callback function are stored within the context to make sure
        of no memory access issues. These data a stored until the free_module_data function
        is called directly by libyang.

        :arg self
            This instance on context
        :arg mod_name:
            The optional YANG module name
        :arg mod_rev:
            The optional YANG module revision
        :arg submod_name:
            The optional YANG submodule name
        :arg submod_rev:
            The optional YANG submodule revision

        :returns:
            Tuple of format string and YANG (sub)module schema
        """
        if self._module_data_clb is None:
            return None
        ret = self._module_data_clb(mod_name, mod_rev, submod_name, submod_rev)
        if ret is None:
            return None
        fmt_str, module_data = ret
        module_data_c = str2c(module_data)
        self._cdata_modules.append(module_data_c)
        return fmt_str, module_data_c

    def set_module_data_clb(
        self,
        clb: Optional[
            Callable[
                [Optional[str], Optional[str], Optional[str], Optional[str]],
                Optional[Tuple[str, str]],
            ]
        ] = None,
    ) -> None:
        """
        Set the callback function, which will be called if libyang context would like to
        load module or submodule, which is not locally available in context path(s).

        :arg self
            This instance on context
        :arg clb:
            The callback function. The expected arguments are:
                mod_name: Module name
                mod_rev: Module revision
                submod_name: Submodule name
                submod_rev: Submodule revision
            The expeted return value is either:
                tuple of:
                    format: The string format of the loaded data
                    data: The YANG (sub)module data as string
                or None in case of error
        """
        self._module_data_clb = clb
        if clb is None:
            lib.ly_ctx_set_module_imp_clb(self._cdata, ffi.NULL, ffi.NULL)
        else:
            lib.ly_ctx_set_module_imp_clb(
                self._cdata, lib.lypy_module_imp_clb, self._cffi_handle
            )


# -------------------------------------------------------------------------------------
class Context:
    __slots__ = (
        "cdata",
        "external_module_loader",
        "__dict__",
    )

    def __init__(
        self,
        search_path: Optional[str] = None,
        disable_searchdirs: bool = False,
        disable_searchdir_cwd: bool = True,
        explicit_compile: Optional[bool] = False,
        leafref_extended: bool = False,
        leafref_linking: bool = False,
        builtin_plugins_only: bool = False,
        all_implemented: bool = False,
        enable_imp_features: bool = False,
        compile_obsolete: bool = False,
        yanglib_path: Optional[str] = None,
        yanglib_fmt: str = "json",
        cdata=None,  # C type: "struct ly_ctx *"
    ):
        if cdata is not None:
            self.cdata = ffi.cast("struct ly_ctx *", cdata)
            self.external_module_loader = ContextExternalModuleLoader(self.cdata)
            return  # already initialized

        options = 0
        if disable_searchdirs:
            options |= lib.LY_CTX_DISABLE_SEARCHDIRS
        if disable_searchdir_cwd:
            options |= lib.LY_CTX_DISABLE_SEARCHDIR_CWD
        if explicit_compile:
            options |= lib.LY_CTX_EXPLICIT_COMPILE
        if leafref_extended:
            options |= lib.LY_CTX_LEAFREF_EXTENDED
        if leafref_linking:
            options |= lib.LY_CTX_LEAFREF_LINKING
        if builtin_plugins_only:
            options |= lib.LY_CTX_BUILTIN_PLUGINS_ONLY
        if all_implemented:
            options |= lib.LY_CTX_ALL_IMPLEMENTED
        if enable_imp_features:
            options |= lib.LY_CTX_ENABLE_IMP_FEATURES
        if compile_obsolete:
            options |= lib.LY_CTX_COMPILE_OBSOLETE
        # force priv parsed
        options |= lib.LY_CTX_SET_PRIV_PARSED

        self.cdata = None
        ctx = ffi.new("struct ly_ctx **")

        search_paths = []
        if "YANGPATH" in os.environ:
            search_paths.extend(os.environ["YANGPATH"].strip(": \t\r\n'\"").split(":"))
        elif "YANG_MODPATH" in os.environ:
            search_paths.extend(
                os.environ["YANG_MODPATH"].strip(": \t\r\n'\"").split(":")
            )
        if search_path:
            search_paths.extend(search_path.strip(": \t\r\n'\"").split(":"))

        search_paths = [path for path in search_paths if os.path.isdir(path)]
        search_path = ":".join(search_paths) if search_paths else None

        if yanglib_path is None:
            options |= lib.LY_CTX_NO_YANGLIBRARY
            if lib.ly_ctx_new(str2c(search_path), options, ctx) != lib.LY_SUCCESS:
                raise self.error("cannot create context")
        else:
            if yanglib_fmt == "json":
                fmt = lib.LYD_JSON
            else:
                fmt = lib.LYD_XML
            print("steweg", search_path, yanglib_path, yanglib_fmt, options)
            ret = lib.ly_ctx_new_ylpath(
                str2c(search_path), str2c(yanglib_path), fmt, options, ctx
            )
            if ret != lib.LY_SUCCESS:
                raise self.error("cannot create context")

        self.cdata = ffi.gc(
            ctx[0],
            lib.ly_ctx_destroy,
        )
        if not self.cdata:
            raise self.error("cannot create context")
        self.external_module_loader = ContextExternalModuleLoader(self.cdata)

    def compile_schema(self):
        ret = lib.ly_ctx_compile(self.cdata)
        if ret != lib.LY_SUCCESS:
            raise self.error("could not compile schema")

    def get_yanglib_data(self, content_id_format=""):
        dnode = ffi.new("struct lyd_node **")
        ret = lib.ly_ctx_get_yanglib_data(self.cdata, dnode, str2c(content_id_format))
        if ret != lib.LY_SUCCESS:
            raise self.error("cannot get yanglib data")
        return DNode.new(self, dnode[0])

    def destroy(self):
        if self.cdata is not None:
            if hasattr(ffi, "release"):
                ffi.release(self.cdata)  # causes ly_ctx_destroy to be called
            self.cdata = None

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.destroy()

    def error(self, msg: str, *args) -> LibyangError:
        if args:
            msg = msg % args

        parts = [msg]
        errors = []

        if self.cdata:
            err = lib.ly_err_first(self.cdata)
            while err:
                m = c2str(err.msg) if err.msg else None
                dp = c2str(err.data_path) if err.data_path else None
                sp = c2str(err.schema_path) if err.schema_path else None
                ln = int(err.line) if err.line else None
                parts.extend(
                    tmpl.format(val)
                    for val, tmpl in [
                        (m, ": {}"),
                        (dp, ": Data path: {}"),
                        (sp, ": Schema path: {}"),
                        (ln, " (line {})"),
                    ]
                    if val is not None
                )
                errors.append(LibyangErrorItem(m, dp, sp, ln))
                err = err.next
            lib.ly_err_clean(self.cdata, ffi.NULL)

        msg = "".join(parts)
        return LibyangError(msg, errors=errors)

    def parse_module(
        self,
        in_data: Union[IO, str],
        in_type: IOType,
        fmt: str = "yang",
        features=None,
    ):
        data = ffi.new("struct ly_in **")
        data_keepalive = []
        ret = data_load(in_type, in_data, data, data_keepalive)
        if ret != lib.LY_SUCCESS:
            raise self.error("failed to read input data")

        feat = ffi.NULL
        if features:
            feat = ffi.new(f"char *[{len(features) + 1}]")
            features = [str2c(i) for i in features]
            for i, val in enumerate(features):
                feat[i] = val
            feat[len(features)] = ffi.NULL

        mod = ffi.new("struct lys_module **")
        fmt = schema_in_format(fmt)
        ret = lib.lys_parse(self.cdata, data[0], fmt, feat, mod)
        lib.ly_in_free(data[0], 0)
        if ret != lib.LY_SUCCESS:
            raise self.error("failed to parse module")

        return Module(self, mod[0])

    def parse_module_file(
        self, fileobj: IO, fmt: str = "yang", features=None
    ) -> Module:
        return self.parse_module(fileobj, IOType.FILE, fmt, features)

    def parse_module_str(self, s: str, fmt: str = "yang", features=None) -> Module:
        return self.parse_module(s, IOType.MEMORY, fmt, features)

    def load_module(
        self,
        name: str,
        revision: Optional[str] = None,
        enabled_features: Sequence[str] = (),
    ) -> Module:
        if self.cdata is None:
            raise RuntimeError("context already destroyed")
        if enabled_features:
            features = tuple([str2c(f) for f in enabled_features] + [ffi.NULL])
        else:
            features = ffi.NULL
        mod = lib.ly_ctx_load_module(self.cdata, str2c(name), str2c(revision), features)
        if mod == ffi.NULL:
            raise self.error("cannot load module")

        return Module(self, mod)

    def get_module(self, name: str) -> Module:
        if self.cdata is None:
            raise RuntimeError("context already destroyed")
        mod = lib.ly_ctx_get_module_latest(self.cdata, str2c(name))
        if mod == ffi.NULL:
            raise self.error("cannot get module")

        return Module(self, mod)

    def find_path(
        self,
        path: str,
        output: bool = False,
        root_node: Optional[SNode] = None,
    ) -> Iterator[SNode]:
        if self.cdata is None:
            raise RuntimeError("context already destroyed")

        if root_node is not None:
            ctx_node = root_node.cdata
        else:
            ctx_node = ffi.NULL

        flags = 0
        if output:
            flags |= lib.LYS_FIND_XP_OUTPUT

        node_set = ffi.new("struct ly_set **")
        if (
            lib.lys_find_xpath(self.cdata, ctx_node, str2c(path), 0, node_set)
            != lib.LY_SUCCESS
        ):
            raise self.error("cannot find path")

        node_set = node_set[0]
        if node_set.count == 0:
            raise self.error("cannot find path")
        try:
            for i in range(node_set.count):
                yield SNode.new(self, node_set.snodes[i])
        finally:
            lib.ly_set_free(node_set, ffi.NULL)

    def find_xpath_atoms(
        self,
        path: str,
        output: bool = False,
        root_node: Optional["libyang.SNode"] = None,
    ) -> Iterator[SNode]:
        if self.cdata is None:
            raise RuntimeError("context already destroyed")

        if root_node is not None:
            ctx_node = root_node.cdata
        else:
            ctx_node = ffi.NULL

        flags = lib.LYS_FIND_XP_OUTPUT if output else 0

        node_set = ffi.new("struct ly_set **")
        if (
            lib.lys_find_xpath_atoms(self.cdata, ctx_node, str2c(path), flags, node_set)
            != lib.LY_SUCCESS
        ):
            raise self.error("cannot find path")

        node_set = node_set[0]
        if node_set.count == 0:
            raise self.error("cannot find path")
        try:
            for i in range(node_set.count):
                yield SNode.new(self, node_set.snodes[i])
        finally:
            lib.ly_set_free(node_set, ffi.NULL)

    def find_jsonpath(
        self,
        path: str,
        root_node: Optional[SNode] = None,
        output: bool = False,
    ) -> Optional[SNode]:
        if root_node is not None:
            ctx_node = root_node.cdata
        else:
            ctx_node = ffi.NULL

        ret = lib.lys_find_path(self.cdata, ctx_node, str2c(path), output)
        if ret == ffi.NULL:
            return None
        return SNode.new(self, ret)

    def create_data_path(
        self,
        path: str,
        parent: Optional[DNode] = None,
        value: Any = None,
        update: bool = True,
        store_only: bool = False,
        rpc_output: bool = False,
        force_return_value: bool = True,
    ) -> Optional[DNode]:
        if self.cdata is None:
            raise RuntimeError("context already destroyed")
        if value is not None:
            if isinstance(value, bool):
                value = str(value).lower()
            elif not isinstance(value, str):
                value = str(value)
        flags = newval_flags(
            update=update, store_only=store_only, rpc_output=rpc_output
        )
        dnode = ffi.new("struct lyd_node **")
        ret = lib.lyd_new_path(
            parent.cdata if parent else ffi.NULL,
            self.cdata,
            str2c(path),
            str2c(value),
            flags,
            dnode,
        )
        dnode = dnode[0]
        if ret != lib.LY_SUCCESS:
            err = lib.ly_err_last(self.cdata)
            if err != ffi.NULL and err.vecode != lib.LYVE_SUCCESS:
                raise self.error("cannot create data path: %s", path)
            lib.ly_err_clean(self.cdata, ffi.NULL)
        if not dnode and not force_return_value:
            return None

        if not dnode and parent:
            # This can happen when path points to an already created leaf and
            # its value does not change.
            # In that case, lookup the existing leaf and return it.
            node_set = ffi.new("struct ly_set **")
            ret = lib.lyd_find_xpath(parent.cdata, str2c(path), node_set)
            if ret != lib.LY_SUCCESS:
                raise self.error("cannot find path: %s", path)

            node_set = node_set[0]
            try:
                if not node_set or not node_set.count:
                    raise self.error("cannot find path: %s", path)
                dnode = node_set.dnodes[0]
            finally:
                lib.ly_set_free(node_set, ffi.NULL)

        if not dnode:
            raise self.error("cannot find created path")

        return DNode.new(self, dnode)

    def parse_op(
        self,
        fmt: str,
        in_type: IOType,
        in_data: Union[IO, str],
        dtype: DataType,
        parent: DNode = None,
        opaq: bool = False,
        strict: bool = False,
    ) -> DNode:
        fmt = data_format(fmt)
        data = ffi.new("struct ly_in **")
        data_keepalive = []
        dtype = data_type(dtype)
        ret = data_load(in_type, in_data, data, data_keepalive)
        if ret != lib.LY_SUCCESS:
            raise self.error("failed to read input data")

        flags = parser_flags(opaq=opaq, strict=strict)
        tree = ffi.new("struct lyd_node **", ffi.NULL)
        op = ffi.new("struct lyd_node **", ffi.NULL)
        par = ffi.new("struct lyd_node **", ffi.NULL)
        if parent is not None:
            par[0] = parent.cdata

        ret = lib.lyd_parse_op(self.cdata, par[0], data[0], fmt, dtype, flags, tree, op)
        lib.ly_in_free(data[0], 0)
        if ret != lib.LY_SUCCESS:
            raise self.error("failed to parse input data")

        return DNode.new(self, op[0])

    def parse_op_mem(
        self,
        fmt: str,
        data: str,
        dtype: DataType = DataType.DATA_YANG,
        parent: DNode = None,
        opaq: bool = False,
        strict: bool = False,
    ):
        return self.parse_op(
            fmt,
            in_type=IOType.MEMORY,
            in_data=data,
            dtype=dtype,
            parent=parent,
            opaq=opaq,
            strict=strict,
        )

    def parse_data(
        self,
        fmt: str,
        in_type: IOType,
        in_data: Union[str, bytes, IO],
        parent: DNode = None,
        no_state: bool = False,
        parse_only: bool = False,
        opaq: bool = False,
        ordered: bool = False,
        strict: bool = False,
        validate_present: bool = False,
        validate_multi_error: bool = False,
        store_only: bool = False,
        json_null: bool = False,
        json_string_datatypes: bool = False,
    ) -> Optional[DNode]:
        if self.cdata is None:
            raise RuntimeError("context already destroyed")
        parser_flgs = parser_flags(
            no_state=no_state,
            parse_only=parse_only,
            opaq=opaq,
            ordered=ordered,
            strict=strict,
            store_only=store_only,
            json_null=json_null,
            json_string_datatypes=json_string_datatypes,
        )
        validation_flgs = validation_flags(
            no_state=no_state,
            validate_present=validate_present,
            validate_multi_error=validate_multi_error,
        )
        fmt = data_format(fmt)
        encode = True
        if fmt == lib.LYD_LYB:
            encode = False
        data = ffi.new("struct ly_in **")
        data_keepalive = []
        ret = data_load(in_type, in_data, data, data_keepalive, encode)
        if ret != lib.LY_SUCCESS:
            raise self.error("failed to read input data")

        parent_cdata = parent.cdata if parent is not None else ffi.NULL
        dnode = ffi.new("struct lyd_node **")
        ret = lib.lyd_parse_data(
            self.cdata, parent_cdata, data[0], fmt, parser_flgs, validation_flgs, dnode
        )
        lib.ly_in_free(data[0], 0)
        if ret != lib.LY_SUCCESS:
            raise self.error("failed to parse data tree")

        dnode = dnode[0]
        if dnode == ffi.NULL:
            return None
        return DNode.new(self, dnode)

    def parse_data_mem(
        self,
        data: Union[str, bytes],
        fmt: str,
        parent: DNode = None,
        no_state: bool = False,
        parse_only: bool = False,
        opaq: bool = False,
        ordered: bool = False,
        strict: bool = False,
        validate_present: bool = False,
        validate_multi_error: bool = False,
        store_only: bool = False,
        json_null: bool = False,
        json_string_datatypes: bool = False,
    ) -> Optional[DNode]:
        return self.parse_data(
            fmt,
            in_type=IOType.MEMORY,
            in_data=data,
            parent=parent,
            no_state=no_state,
            parse_only=parse_only,
            opaq=opaq,
            ordered=ordered,
            strict=strict,
            validate_present=validate_present,
            validate_multi_error=validate_multi_error,
            store_only=store_only,
            json_null=json_null,
            json_string_datatypes=json_string_datatypes,
        )

    def parse_data_file(
        self,
        fileobj: IO,
        fmt: str,
        parent: DNode = None,
        no_state: bool = False,
        parse_only: bool = False,
        opaq: bool = False,
        ordered: bool = False,
        strict: bool = False,
        validate_present: bool = False,
        validate_multi_error: bool = False,
        store_only: bool = False,
        json_null: bool = False,
        json_string_datatypes: bool = False,
    ) -> Optional[DNode]:
        return self.parse_data(
            fmt,
            in_type=IOType.FD,
            in_data=fileobj,
            parent=parent,
            no_state=no_state,
            parse_only=parse_only,
            opaq=opaq,
            ordered=ordered,
            strict=strict,
            validate_present=validate_present,
            validate_multi_error=validate_multi_error,
            store_only=store_only,
            json_null=json_null,
            json_string_datatypes=json_string_datatypes,
        )

    def find_leafref_path_target_paths(self, leafref_path: str) -> List[str]:
        """
        Fetch all leafref targets of the specified path

        This is an enhanced version of lysc_node_lref_target() which will return
        a set of leafref target paths retrieved from the specified schema path.
        While lysc_node_lref_target() will only work on nodetype of LYS_LEAF and
        LYS_LEAFLIST this function will also evaluate other datatypes that may
        contain leafrefs such as LYS_UNION.  This does not, however, search for
        children with leafref targets.

        :arg self
            This instance on context
        :arg leafref_path:
            Path to node to search for leafref targets
        :returns List of target paths that the leafrefs of the specified node
                 point to.
        """
        if self.cdata is None:
            raise RuntimeError("context already destroyed")
        if leafref_path is None:
            raise RuntimeError("leafref_path must be defined")

        out = []

        node = lib.lys_find_path(self.cdata, ffi.NULL, str2c(leafref_path), 0)
        if node == ffi.NULL:
            raise self.error("leafref_path not found")

        node_set = ffi.new("struct ly_set **")
        if (
            lib.lysc_node_lref_targets(node, node_set) != lib.LY_SUCCESS
            or node_set[0] == ffi.NULL
            or node_set[0].count == 0
        ):
            raise self.error("leafref_path does not contain any leafref targets")

        node_set = node_set[0]
        for i in range(node_set.count):
            path = lib.lysc_path(node_set.snodes[i], lib.LYSC_PATH_DATA, ffi.NULL, 0)
            out.append(c2str(path))
            lib.free(path)

        lib.ly_set_free(node_set, ffi.NULL)

        return out

    def find_backlinks_paths(
        self, match_path: str = None, match_ancestors: bool = False
    ) -> List[str]:
        """
        Search entire schema for nodes that contain leafrefs and return as a
        list of schema node paths.

        Perform a complete scan of the schema tree looking for nodes that
        contain leafref entries. When a node contains a leafref entry, and
        match_path is specified, determine if reference points to match_path,
        if so add the node's path to returned list.  If no match_path is
        specified, the node containing the leafref is always added to the
        returned set.  When match_ancestors is true, will evaluate if match_path
        is self or an ansestor of self.

        This does not return the leafref targets, but the actual node that
        contains a leafref.

        :arg self
            This instance on context
        :arg match_path:
            Target path to use for matching
        :arg match_ancestors:
            Whether match_path is a base ancestor or an exact node
        :returns List of paths.  Exception of match_path is not found or if no
                 backlinks are found.
        """
        if self.cdata is None:
            raise RuntimeError("context already destroyed")
        out = []

        match_node = ffi.NULL
        if match_path is not None and match_path == "/" or match_path == "":
            match_path = None

        if match_path:
            match_node = lib.lys_find_path(self.cdata, ffi.NULL, str2c(match_path), 0)
            if match_node == ffi.NULL:
                raise self.error("match_path not found")

        node_set = ffi.new("struct ly_set **")
        if (
            lib.lysc_node_lref_backlinks(
                self.cdata, match_node, match_ancestors, node_set
            )
            != lib.LY_SUCCESS
            or node_set[0] == ffi.NULL
            or node_set[0].count == 0
        ):
            raise self.error("backlinks not found")

        node_set = node_set[0]
        for i in range(node_set.count):
            path = lib.lysc_path(node_set.snodes[i], lib.LYSC_PATH_DATA, ffi.NULL, 0)
            out.append(c2str(path))
            lib.free(path)

        lib.ly_set_free(node_set, ffi.NULL)

        return out

    def __iter__(self) -> Iterator[Module]:
        """
        Return an iterator that yields all implemented modules from the context
        """
        if self.cdata is None:
            raise RuntimeError("context already destroyed")
        idx = ffi.new("uint32_t *")
        mod = lib.ly_ctx_get_module_iter(self.cdata, idx)
        while mod:
            yield Module(self, mod)
            mod = lib.ly_ctx_get_module_iter(self.cdata, idx)

    def add_to_dict(self, orig_str: str) -> Any:
        cstr = ffi.new("char **")
        ret = lib.lydict_insert(self.cdata, str2c(orig_str), 0, cstr)
        if ret != lib.LY_SUCCESS:
            raise LibyangError("Unable to insert string into context dictionary")
        return cstr[0]

    def remove_from_dict(self, orig_str: str) -> None:
        lib.lydict_remove(self.cdata, str2c(orig_str))



# Copyright (c) 2025, LabN Consulting, L.L.C.
# SPDX-License-Identifier: MIT

import logging
import os
import sys
import unittest

from libyang import Context, LibyangError, configure_logging, temp_log_options


YANG_DIR = os.path.join(os.path.dirname(__file__), "yang")


class LogTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        configure_logging(False, logging.INFO)

    def tearDown(self):
        if self.ctx is not None:
            self.ctx.destroy()
        self.ctx = None

    def _cause_log(self):
        try:
            assert self.ctx is not None
            _ = self.ctx.parse_data_mem("bad", fmt="xml")
        except LibyangError:
            pass

    @unittest.skipIf(sys.version_info < (3, 10), "Test requires Python 3.10+")
    def test_configure_logging(self):
        """Test configure_logging API."""
        with self.assertNoLogs("libyang", level="ERROR"):
            self._cause_log()

        configure_logging(True, logging.INFO)
        with self.assertLogs("libyang", level="ERROR"):
            self._cause_log()

    @unittest.skipIf(sys.version_info < (3, 10), "Test requires Python 3.10+")
    def test_with_temp_log(self):
        """Test configure_logging API."""
        configure_logging(True, logging.INFO)

        with self.assertLogs("libyang", level="ERROR"):
            self._cause_log()

        with self.assertNoLogs("libyang", level="ERROR"):
            with temp_log_options(0):
                self._cause_log()

        with self.assertLogs("libyang", level="ERROR"):
            self._cause_log()



# Copyright (c) 2019 6WIND
# SPDX-License-Identifier: MIT

from typing import Any, Callable, Iterator, Optional

from .context import Context
from .schema import SContainer, SLeaf, SLeafList, SList, SNode, SNotif, SRpc, SRpcInOut


# -------------------------------------------------------------------------------------
def schema_diff(
    ctx_old: Context,
    ctx_new: Context,
    exclude_node_cb: Optional[Callable[[SNode], bool]] = None,
    use_data_path: bool = False,
) -> Iterator["SNodeDiff"]:
    """
    Compare two libyang Context objects, for a given set of paths and return all
    differences.

    :arg ctx_old:
        The first context.
    :arg ctx_new:
        The second context.
    :arg exclude_node_cb:
        Optional user callback that will be called with each node that is found in each
        context. If the callback returns a "trueish" value, the node will be excluded
        from the diff (as well as all its children).
    :arg use_data_path:
        Use data path instead of schema path to compare the nodes. Using data path
        ignores choices and cases.

    :return:
        An iterator that yield `SNodeDiff` objects.
    """
    if exclude_node_cb is None:
        # pylint: disable=unnecessary-lambda-assignment
        exclude_node_cb = lambda n: False

    def flatten(node, dic):
        """
        Flatten a node and all its children into a dict (indexed by their schema xpath).
        This function is recursive.
        """
        if exclude_node_cb(node):
            return
        if use_data_path:
            path = node.data_path()
        else:
            path = node.schema_path()
        dic[path] = node
        if isinstance(node, (SContainer, SList, SNotif, SRpc, SRpcInOut)):
            for child in node:
                flatten(child, dic)

    old_dict = {}
    new_dict = {}

    for module in ctx_old:
        for node in module:
            flatten(node, old_dict)
    for module in ctx_new:
        for node in module:
            flatten(node, new_dict)

    diffs = {}
    old_paths = frozenset(old_dict.keys())
    new_paths = frozenset(new_dict.keys())
    for path in old_paths - new_paths:
        diffs[path] = [SNodeRemoved(old_dict[path])]
    for path in new_paths - old_paths:
        diffs[path] = [SNodeAdded(new_dict[path])]
    for path in old_paths & new_paths:
        old = old_dict[path]
        new = new_dict[path]
        diffs[path] = snode_changes(old, new)

    for path in sorted(diffs.keys()):
        yield from diffs[path]


# -------------------------------------------------------------------------------------
class SNodeDiff:
    __slots__ = ("__dict__",)


# -------------------------------------------------------------------------------------
class SNodeRemoved(SNodeDiff):
    __slots__ = ("node",)

    def __init__(self, node: SNode):
        self.node = node

    def __str__(self):
        return "-%s: removed status=%s %s" % (
            self.node.schema_path(),
            self.node.status(),
            self.node.keyword(),
        )


# -------------------------------------------------------------------------------------
class SNodeAdded(SNodeDiff):
    __slots__ = ("node",)

    def __init__(self, node: SNode):
        self.node = node

    def __str__(self):
        return "+%s: added %s" % (self.node.schema_path(), self.node.keyword())


# -------------------------------------------------------------------------------------
class SNodeAttributeChanged(SNodeDiff):
    __slots__ = ("old", "new", "value")

    def __init__(self, old: SNode, new: SNode, value: Any = None):
        self.old = old
        self.new = new
        self.value = value

    def __str__(self):
        if self.__class__.__name__.endswith("Added"):
            sign = "+"
        else:
            sign = "-"
        s = "%s%s: %s" % (sign, self.new.schema_path(), self.__class__.__name__)
        if self.value is not None:
            str_val = str(self.value).replace('"', '\\"').replace("\n", "\\n")
            s += ' "%s"' % str_val
        return s


# -------------------------------------------------------------------------------------
class BaseTypeRemoved(SNodeAttributeChanged):
    pass


class BaseTypeAdded(SNodeAttributeChanged):
    pass


class BitRemoved(SNodeAttributeChanged):
    pass


class BitAdded(SNodeAttributeChanged):
    pass


class BitStatusRemoved(SNodeAttributeChanged):
    pass


class BitStatusAdded(SNodeAttributeChanged):
    pass


class ConfigFalseRemoved(SNodeAttributeChanged):
    pass


class ConfigFalseAdded(SNodeAttributeChanged):
    pass


class DefaultRemoved(SNodeAttributeChanged):
    pass


class DefaultAdded(SNodeAttributeChanged):
    pass


class DescriptionRemoved(SNodeAttributeChanged):
    pass


class DescriptionAdded(SNodeAttributeChanged):
    pass


class EnumRemoved(SNodeAttributeChanged):
    pass


class EnumAdded(SNodeAttributeChanged):
    pass


class EnumStatusRemoved(SNodeAttributeChanged):
    pass


class EnumStatusAdded(SNodeAttributeChanged):
    pass


class ExtensionRemoved(SNodeAttributeChanged):
    pass


class ExtensionAdded(SNodeAttributeChanged):
    pass


class KeyRemoved(SNodeAttributeChanged):
    pass


class KeyAdded(SNodeAttributeChanged):
    pass


class MandatoryRemoved(SNodeAttributeChanged):
    pass


class MandatoryAdded(SNodeAttributeChanged):
    pass


class MustRemoved(SNodeAttributeChanged):
    pass


class MustAdded(SNodeAttributeChanged):
    pass


class NodeTypeAdded(SNodeAttributeChanged):
    pass


class NodeTypeRemoved(SNodeAttributeChanged):
    pass


class RangeRemoved(SNodeAttributeChanged):
    pass


class RangeAdded(SNodeAttributeChanged):
    pass


class OrderedByUserRemoved(SNodeAttributeChanged):
    pass


class OrderedByUserAdded(SNodeAttributeChanged):
    pass


class PresenceRemoved(SNodeAttributeChanged):
    pass


class PresenceAdded(SNodeAttributeChanged):
    pass


class StatusRemoved(SNodeAttributeChanged):
    pass


class StatusAdded(SNodeAttributeChanged):
    pass


class LengthRemoved(SNodeAttributeChanged):
    pass


class LengthAdded(SNodeAttributeChanged):
    pass


class PatternRemoved(SNodeAttributeChanged):
    pass


class PatternAdded(SNodeAttributeChanged):
    pass


class UnitsRemoved(SNodeAttributeChanged):
    pass


class UnitsAdded(SNodeAttributeChanged):
    pass


# -------------------------------------------------------------------------------------
def snode_changes(old: SNode, new: SNode) -> Iterator[SNodeDiff]:
    if old.nodetype() != new.nodetype():
        yield NodeTypeRemoved(old, new, old.keyword())
        yield NodeTypeAdded(old, new, new.keyword())

    if old.description() != new.description():
        if old.description() is not None:
            yield DescriptionRemoved(old, new, old.description())
        if new.description() is not None:
            yield DescriptionAdded(old, new, new.description())

    if old.mandatory() and not new.mandatory():
        yield MandatoryRemoved(old, new)
    elif not old.mandatory() and new.mandatory():
        yield MandatoryAdded(old, new)

    if old.status() != new.status():
        yield StatusRemoved(old, new, old.status())
        yield StatusAdded(old, new, new.status())

    if old.config_false() and not new.config_false():
        yield ConfigFalseRemoved(old, new)
    elif not old.config_false() and new.config_false():
        yield ConfigFalseAdded(old, new)

    old_musts = frozenset(old.must_conditions())
    new_musts = frozenset(new.must_conditions())
    for m in old_musts - new_musts:
        yield MustRemoved(old, new, m)
    for m in new_musts - old_musts:
        yield MustAdded(old, new, m)

    old_exts = {(e.module().name(), e.name()): e for e in old.extensions()}
    new_exts = {(e.module().name(), e.name()): e for e in new.extensions()}
    old_ext_keys = frozenset(old_exts.keys())
    new_ext_keys = frozenset(new_exts.keys())

    for k in old_ext_keys - new_ext_keys:
        yield ExtensionRemoved(old, new, old_exts[k])
    for k in new_ext_keys - old_ext_keys:
        yield ExtensionAdded(old, new, new_exts[k])
    for k in old_ext_keys & new_ext_keys:
        if old_exts[k].argument() != new_exts[k].argument():
            yield ExtensionRemoved(old, new, old_exts[k])
            yield ExtensionAdded(old, new, new_exts[k])

    if (isinstance(old, SLeaf) and isinstance(new, SLeaf)) or (
        isinstance(old, SLeafList) and isinstance(new, SLeafList)
    ):
        old_bases = set(old.type().basenames())
        new_bases = set(new.type().basenames())
        for b in old_bases - new_bases:
            yield BaseTypeRemoved(old, new, b)
        for b in new_bases - old_bases:
            yield BaseTypeAdded(old, new, b)

        if old.units() != new.units():
            if old.units() is not None:
                yield UnitsRemoved(old, new, old.units())
            if new.units() is not None:
                yield UnitsAdded(old, new, new.units())

        old_lengths = set(old.type().all_lengths())
        new_lengths = set(new.type().all_lengths())
        for l in old_lengths - new_lengths:
            yield LengthRemoved(old, new, l)
        for l in new_lengths - old_lengths:
            yield LengthAdded(old, new, l)

        # Multiple "pattern" statements on a single type are ANDed together (i.e. they
        # must all match). However, when a leaf type is an union of multiple string
        # typedefs with individual "pattern" statements, the patterns are ORed together
        # (i.e. one of them must match).
        #
        # This is not handled properly here as we flatten all patterns in a single set
        # and consider there is a difference if we remove/add one of them.
        #
        # The difference does not hold any information about which union branch it
        # relates to. This is way too complex.
        old_patterns = set(old.type().all_patterns())
        new_patterns = set(new.type().all_patterns())
        for p in old_patterns - new_patterns:
            yield PatternRemoved(old, new, p)
        for p in new_patterns - old_patterns:
            yield PatternAdded(old, new, p)

        old_ranges = set(old.type().all_ranges())
        new_ranges = set(new.type().all_ranges())
        for r in old_ranges - new_ranges:
            yield RangeRemoved(old, new, r)
        for r in new_ranges - old_ranges:
            yield RangeAdded(old, new, r)

        old_enums = {e.name(): e for e in old.type().all_enums()}
        new_enums = {e.name(): e for e in new.type().all_enums()}
        for e in old_enums.keys() - new_enums.keys():
            yield EnumRemoved(old, new, old_enums[e])
        for e in new_enums.keys() - old_enums.keys():
            yield EnumAdded(old, new, new_enums[e])
        for e in new_enums.keys() & old_enums.keys():
            o = old_enums[e]
            n = new_enums[e]
            if o.status() != n.status():
                yield EnumStatusRemoved(old, new, (o.name(), o.status()))
                yield EnumStatusAdded(old, new, (n.name(), n.status()))

        old_bits = {b.name(): b for b in old.type().all_bits()}
        new_bits = {b.name(): b for b in new.type().all_bits()}
        for b in old_bits.keys() - new_bits.keys():
            yield BitRemoved(old, new, old_bits[b])
        for b in new_bits.keys() - old_bits.keys():
            yield BitAdded(old, new, new_bits[b])
        for b in new_bits.keys() & old_bits.keys():
            o = old_bits[b]
            n = new_bits[b]
            if o.status() != n.status():
                yield BitStatusRemoved(old, new, (o.name(), o.status()))
                yield BitStatusAdded(old, new, (n.name(), n.status()))

    if isinstance(old, SLeaf) and isinstance(new, SLeaf):
        if old.default() != new.default():
            if old.default() is not None:
                yield DefaultRemoved(old, new, old.default())
            if new.default() is not None:
                yield DefaultAdded(old, new, new.default())

    elif isinstance(old, SLeafList) and isinstance(new, SLeafList):
        old_defaults = frozenset(old.defaults())
        new_defaults = frozenset(new.defaults())
        for d in old_defaults - new_defaults:
            yield DefaultRemoved(old, new, d)
        for d in new_defaults - old_defaults:
            yield DefaultAdded(old, new, d)
        if old.ordered() and not new.ordered():
            yield OrderedByUserRemoved(old, new)
        elif not old.ordered() and new.ordered():
            yield OrderedByUserAdded(old, new)

    elif isinstance(old, SContainer) and isinstance(new, SContainer):
        if old.presence() != new.presence():
            if old.presence() is not None:
                yield PresenceRemoved(old, new, old.presence())
            if new.presence() is not None:
                yield PresenceAdded(old, new, new.presence())

    elif isinstance(old, SList) and isinstance(new, SList):
        old_keys = frozenset(k.name() for k in old.keys())
        new_keys = frozenset(k.name() for k in new.keys())
        for k in old_keys - new_keys:
            yield KeyRemoved(old, new, k)
        for k in new_keys - old_keys:
            yield KeyAdded(old, new, k)
        if old.ordered() and not new.ordered():
            yield OrderedByUserRemoved(old, new)
        elif not old.ordered() and new.ordered():
            yield OrderedByUserAdded(old, new)



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: MIT

import gc
import json
import os
import unittest
from unittest import mock
from unittest.mock import patch

from _libyang import lib
from libyang import (
    Context,
    DAnyxml,
    DataType,
    DContainer,
    DLeaf,
    DList,
    DNode,
    DNodeAttrs,
    DNotif,
    DRpc,
    IOType,
    LibyangError,
    Module,
)
from libyang.data import dict_to_dnode


YANG_DIR = os.path.join(os.path.dirname(__file__), "yang")


# -------------------------------------------------------------------------------------
class DataTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR, compile_obsolete=True)
        modules = [
            self.ctx.load_module("ietf-netconf"),
            self.ctx.load_module("yolo-system"),
            self.ctx.load_module("yolo-nodetypes"),
        ]

        for mod in modules:
            mod.feature_enable_all()

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    JSON_CONFIG = """{
  "yolo-system:conf": {
    "hostname": "foo",
    "url": [
      {
        "proto": "http",
        "host": "foobar.com",
        "port": 8080,
        "path": "/index.html",
        "enabled": true
      },
      {
        "proto": "https",
        "host": "github.com",
        "path": "/CESNET/libyang-python",
        "enabled": false
      }
    ],
    "number": [
      1000,
      2000,
      3000
    ],
    "speed": 1234
  }
}
"""

    def test_data_parse_config_json(self):
        dnode = self.ctx.parse_data_mem(
            self.JSON_CONFIG, "json", no_state=True, ordered=True
        )
        self.assertIsInstance(dnode, DContainer)
        try:
            j = dnode.print_mem("json", with_siblings=True)
            self.assertEqual(j, self.JSON_CONFIG)
        finally:
            dnode.free()

    JSON_CONFIG_WITH_STATE = """{
  "yolo-system:state": {
    "speed": 4321
  },
  "yolo-system:conf": {
    "hostname": "foo",
    "url": [
      {
        "proto": "http",
        "host": "foobar.com",
        "port": 8080,
        "path": "/index.html",
        "enabled": true
      },
      {
        "proto": "https",
        "host": "github.com",
        "path": "/CESNET/libyang-python",
        "enabled": false
      }
    ],
    "number": [
      1000,
      2000,
      3000
    ],
    "speed": 1234
  }
}
"""

    def test_data_parse_config_json_without_yang_lib(self):
        dnode = self.ctx.parse_data_mem(self.JSON_CONFIG, "json")
        self.assertIsInstance(dnode, DContainer)
        try:
            j = dnode.print_mem("json", with_siblings=True)
            self.assertEqual(j, self.JSON_CONFIG_WITH_STATE)
        finally:
            dnode.free()

    JSON_CONFIG_ADD_LIST_ITEM = """{
  "yolo-system:conf": {
    "hostname": "foo",
    "url": [
      {
        "proto": "http",
        "host": "barfoo.com",
        "path": "/barfoo/index.html"
      },
      {
        "proto": "http",
        "host": "foobar.com",
        "port": 8080,
        "path": "/index.html",
        "enabled": true
      },
      {
        "proto": "https",
        "host": "github.com",
        "path": "/CESNET/libyang-python",
        "enabled": false
      }
    ],
    "number": [
      1000,
      2000,
      3000
    ],
    "speed": 1234
  }
}
"""

    def test_data_add_path(self):
        dnode = self.ctx.parse_data_mem(self.JSON_CONFIG, "json", no_state=True)
        dnode.new_path(
            '/yolo-system:conf/url[host="barfoo.com"][proto="http"]/path',
            "/barfoo/index.html",
        )
        self.assertIsInstance(dnode, DContainer)
        try:
            j = dnode.print_mem("json", with_siblings=True)
            self.assertEqual(j, self.JSON_CONFIG_ADD_LIST_ITEM)
        finally:
            dnode.free()

    JSON_CONFIG_FILE = os.path.join(os.path.dirname(__file__), "data/config.json")

    def test_data_parse_config_json_file(self):
        with open(self.JSON_CONFIG_FILE, encoding="utf-8") as f:
            dnode = self.ctx.parse_data_file(f, "json", no_state=True)
        self.assertIsInstance(dnode, DContainer)
        dnode.free()

        with open(self.JSON_CONFIG_FILE, encoding="utf-8") as f:
            dnode = self.ctx.parse_data(
                "json", in_data=f, in_type=IOType.FILE, no_state=True
            )
        self.assertIsInstance(dnode, DContainer)
        dnode.free()

        dnode = self.ctx.parse_data(
            "json",
            in_data=self.JSON_CONFIG_FILE,
            in_type=IOType.FILEPATH,
            no_state=True,
        )
        self.assertIsInstance(dnode, DContainer)
        dnode.free()

    JSON_STATE = """{
  "yolo-system:state": {
    "hostname": "foo",
    "url": [
      {
        "proto": "https",
        "host": "github.com",
        "path": "/CESNET/libyang-python",
        "enabled": false
      },
      {
        "proto": "http",
        "host": "foobar.com",
        "port": 8080,
        "path": "/index.html",
        "enabled": true
      }
    ],
    "number": [
      1000,
      2000,
      3000
    ],
    "speed": 1234
  }
}
"""

    def test_data_parse_state_json(self):
        dnode = self.ctx.parse_data_mem(self.JSON_STATE, "json", validate_present=True)
        self.assertIsInstance(dnode, DContainer)
        try:
            j = dnode.print_mem("json", with_siblings=True)
            self.assertEqual(j, self.JSON_STATE)
        finally:
            dnode.free()

    XML_CONFIG = """<conf xmlns="urn:yang:yolo:system">
  <hostname>foo</hostname>
  <url>
    <proto>https</proto>
    <host>github.com</host>
    <path>/CESNET/libyang-python</path>
    <enabled>false</enabled>
  </url>
  <url>
    <proto>http</proto>
    <host>foobar.com</host>
    <port>8080</port>
    <path>/index.html</path>
    <enabled>true</enabled>
  </url>
  <number>1000</number>
  <number>2000</number>
  <number>3000</number>
  <speed>1234</speed>
</conf>
"""

    def test_data_parse_config_xml(self):
        dnode = self.ctx.parse_data_mem(
            self.XML_CONFIG, "xml", validate_present=True, ordered=True
        )
        self.assertIsInstance(dnode, DContainer)
        try:
            xml = dnode.print_mem("xml", with_siblings=True, trim_default_values=True)
            self.assertEqual(xml, self.XML_CONFIG)
        finally:
            dnode.free()

    XML_CONFIG_MULTI_ERROR = """<conf xmlns="urn:yang:yolo:system">
  <hostname>foo</hostname>
  <url>
    <proto>https</proto>
    <path>/CESNET/libyang-python</path>
    <enabled>abcd</enabled>
  </url>
  <number>2000</number>
</conf>
"""

    def test_data_parse_config_xml_multi_error(self):
        with self.assertRaises(Exception) as cm:
            self.ctx.parse_data_mem(
                self.XML_CONFIG_MULTI_ERROR,
                "xml",
                validate_present=True,
                validate_multi_error=True,
            )
        self.assertEqual(
            str(cm.exception),
            'failed to parse data tree: Invalid boolean value "abcd".: '
            "Data path: /yolo-system:conf/url[proto='https']/enabled (line 6): "
            'List instance is missing its key "host".: '
            "Data path: /yolo-system:conf/url[proto='https'] (line 7)",
        )

        first = cm.exception.errors[0]
        self.assertEqual(first.msg, 'Invalid boolean value "abcd".')
        self.assertEqual(
            first.data_path, "/yolo-system:conf/url[proto='https']/enabled"
        )
        self.assertEqual(first.line, 6)

        second = cm.exception.errors[1]
        self.assertEqual(second.msg, 'List instance is missing its key "host".')
        self.assertEqual(second.data_path, "/yolo-system:conf/url[proto='https']")
        self.assertEqual(second.line, 7)

    XML_STATE = """<state xmlns="urn:yang:yolo:system">
  <hostname>foo</hostname>
  <url>
    <proto>https</proto>
    <host>github.com</host>
    <path>/CESNET/libyang-python</path>
    <enabled>false</enabled>
  </url>
  <url>
    <proto>http</proto>
    <host>foobar.com</host>
    <port>8080</port>
    <path>/index.html</path>
    <enabled>true</enabled>
  </url>
  <number>1000</number>
  <number>2000</number>
  <number>3000</number>
  <speed>1234</speed>
</state>
"""

    def test_data_parse_data_xml(self):
        dnode = self.ctx.parse_data_mem(self.XML_STATE, "xml", validate_present=True)
        self.assertIsInstance(dnode, DContainer)
        try:
            xml = dnode.print("xml", out_type=IOType.MEMORY, with_siblings=True)
            self.assertEqual(xml, self.XML_STATE)
        finally:
            dnode.free()

    XML_NETCONF_IN = """<rpc xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
            <edit-config>
              <target>
                <running/>
              </target>
              <config xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
                <conf xmlns="urn:yang:yolo:system">
                    <hostname-ref>notdefined</hostname-ref>
                </conf>
              </config>
            </edit-config>
            </rpc>
            """

    XML_NETCONF_OUT = """<edit-config xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
  <target>
    <running/>
  </target>
  <config>
    <conf xmlns="urn:yang:yolo:system">
      <hostname-ref>notdefined</hostname-ref>
    </conf>
  </config>
</edit-config>
"""

    def test_data_parse_netconf(self):
        dnode = self.ctx.parse_op_mem("xml", self.XML_NETCONF_IN, DataType.RPC_NETCONF)
        self.assertIsInstance(dnode, DContainer)
        try:
            xml = dnode.print("xml", out_type=IOType.MEMORY)
            self.assertEqual(xml, self.XML_NETCONF_OUT)
        finally:
            dnode.free()

    ANYMXML = """<format-disk xmlns="urn:yang:yolo:system">
      <html-info>
        <p xmlns="http://www.w3.org/1999/xhtml">
        </p>
      </html-info>
</format-disk>
"""

    def test_data_parse_anyxml(self):
        dnode = self.ctx.parse_op_mem("xml", self.ANYMXML, dtype=DataType.RPC_YANG)
        dnode = dnode.find_path("/yolo-system:format-disk/html-info")
        self.assertIsInstance(dnode, DAnyxml)

    def test_data_create_paths(self):
        state = self.ctx.create_data_path("/yolo-system:state")
        try:
            state.create_path("hostname", "foo")
            state.create_path("speed", 1234)
            state.create_path("number", 1000)
            state.create_path("number", 2000)
            state.create_path("number", 3000)
            u = state.create_path('url[proto="https"][host="github.com"]')
            u.create_path("path", "/CESNET/libyang-python")
            u.create_path("enabled", False)
            u = state.create_path('url[proto="http"][host="foobar.com"]')
            u.create_path("port", 8080)
            u.create_path("path", "/index.html")
            u.create_path("enabled", True)
            self.assertEqual(state.print_mem("json"), self.JSON_STATE)
        finally:
            state.free()

    def test_data_create_invalid_type(self):
        s = self.ctx.create_data_path("/yolo-system:state")
        try:
            with self.assertRaises(LibyangError):
                s.create_path("speed", 1234000000000000000000000000)
        finally:
            s.free()

    def test_data_create_invalid_regexp(self):
        s = self.ctx.create_data_path("/yolo-system:state")
        try:
            with self.assertRaises(LibyangError):
                s.create_path("hostname", "INVALID.HOST")
        finally:
            s.free()

    DICT_CONFIG = {
        "conf": {
            "hostname": "foo",
            "speed": 1234,
            "number": [1000, 2000, 3000],
            "url": [
                {
                    "enabled": False,
                    "path": "/CESNET/libyang-python",
                    "host": "github.com",
                    "proto": "https",
                },
                {
                    "port": 8080,
                    "proto": "http",
                    "path": "/index.html",
                    "enabled": True,
                    "host": "foobar.com",
                },
            ],
        }
    }

    def test_data_to_dict_config(self):
        dnode = self.ctx.parse_data_mem(self.JSON_CONFIG, "json", validate_present=True)
        self.assertIsInstance(dnode, DContainer)
        try:
            dic = dnode.print_dict()
        finally:
            dnode.free()
        self.assertEqual(dic, self.DICT_CONFIG)

    def test_data_to_dict_rpc_input(self):
        in_data = '{"yolo-system:format-disk": {"disk": "/dev/sda"}}'
        dnode = self.ctx.parse_op_mem("json", in_data, DataType.RPC_YANG)
        self.assertIsInstance(dnode, DRpc)
        try:
            dic = dnode.print_dict()
        finally:
            dnode.free()
        self.assertEqual(dic, {"format-disk": {"disk": "/dev/sda"}})

    def test_data_from_dict_module(self):
        module = self.ctx.get_module("yolo-system")
        dnode = module.parse_data_dict(
            self.DICT_CONFIG, strict=True, validate_present=True
        )
        self.assertIsInstance(dnode, DContainer)
        try:
            j = dnode.print_mem("json")
        finally:
            dnode.free()
        self.assertEqual(json.loads(j), json.loads(self.JSON_CONFIG))

    def test_data_from_dict_module_free_func(self):
        module = self.ctx.get_module("yolo-system")

        free_func = mock.Mock()
        dnode = module.parse_data_dict(
            self.DICT_CONFIG, strict=True, validate_present=True
        )
        dnode.free_func = free_func
        self.assertIsInstance(dnode, DContainer)
        try:
            j = dnode.print_mem("json")
        finally:
            dnode.free()
        self.assertEqual(json.loads(j), json.loads(self.JSON_CONFIG))
        free_func.assert_called()

    DICT_CONFIG_WITH_PREFIX = {
        "yolo-system:conf": {
            "hostname": "foo",
            "speed": 1234,
            "number": [1000, 2000, 3000],
            "url": [
                {
                    "enabled": False,
                    "path": "/CESNET/libyang-python",
                    "host": "github.com",
                    "proto": "https",
                },
                {
                    "port": 8080,
                    "proto": "http",
                    "path": "/index.html",
                    "enabled": True,
                    "host": "foobar.com",
                },
            ],
        }
    }

    def test_data_from_dict_module_with_prefix(self):
        self.maxDiff = None
        module = self.ctx.get_module("yolo-system")
        dnode = module.parse_data_dict(
            self.DICT_CONFIG_WITH_PREFIX, strict=True, validate_present=True
        )
        self.assertIsInstance(dnode, DContainer)
        try:
            j = dnode.print_mem("json")
        finally:
            dnode.free()
        self.assertEqual(json.loads(j), json.loads(self.JSON_CONFIG))

    def test_data_from_dict_invalid(self):
        module = self.ctx.get_module("yolo-system")

        created = []
        freed = []

        class FakeLib:
            def __init__(self, orig):
                self.orig = orig

            def lyd_new_inner(self, *args):
                ret = self.orig.lyd_new_inner(*args)
                c = args[4][0]
                if ret == lib.LY_SUCCESS:
                    created.append(c)
                return ret

            def lyd_new_term(self, *args):
                ret = self.orig.lyd_new_term(*args)
                c = args[5][0]
                if ret == lib.LY_SUCCESS:
                    created.append(c)
                return ret

            def lyd_new_list(self, *args):
                ret = self.orig.lyd_new_list(*args)
                c = args[4][0]
                if ret == lib.LY_SUCCESS:
                    created.append(c)
                return ret

            def lyd_free_tree(self, dnode):
                freed.append(dnode)
                self.orig.lyd_free_tree(dnode)

            def __getattr__(self, name):
                return getattr(self.orig, name)

        fake_lib = FakeLib(lib)
        root = module.parse_data_dict(
            {"conf": {"hostname": "foo", "speed": 1234, "number": [1000, 2000, 3000]}},
            strict=True,
            validate_present=True,
        )
        invalid_dict = {
            "url": [
                {
                    "host": "github.com",
                    "proto": "https",
                    "enabled": False,
                    "path": "/CESNET/libyang-python",
                },
                {
                    "proto": "http",
                    "path": "/index.html",
                    "enabled": True,
                    "host": "foobar.com",
                    "port": "INVALID.PORT",
                },
            ]
        }

        try:
            with patch("libyang.data.lib", fake_lib):
                with self.assertRaises(LibyangError):
                    root.merge_data_dict(
                        invalid_dict, strict=True, validate_present=True
                    )

            self.assertGreater(len(created), 0)
            self.assertGreater(len(freed), 0)
            self.assertEqual(freed, list(reversed(created)))
        finally:
            root.free()

    def test_data_from_dict_container(self):
        dnode = self.ctx.create_data_path("/yolo-system:conf")
        self.assertIsInstance(dnode, DContainer)
        subtree = dnode.merge_data_dict(
            self.DICT_CONFIG["conf"], strict=True, validate_present=True
        )
        # make sure subtree validation is forbidden
        with self.assertRaises(LibyangError):
            subtree.validate(validate_present=True)
        try:
            dnode.validate(validate_present=True)
            j = dnode.print_mem("json")
        finally:
            dnode.free()
        self.assertEqual(json.loads(j), json.loads(self.JSON_CONFIG))

    def test_data_from_dict_leaf(self):
        dnode = self.ctx.create_data_path("/yolo-system:state")
        self.assertIsInstance(dnode, DContainer)
        dnode.merge_data_dict(
            {"hostname": "foo"}, strict=True, validate=True, validate_present=True
        )
        try:
            j = dnode.print_mem("json", pretty=False, trim_default_values=True)
        finally:
            dnode.free()
        self.assertEqual(j, '{"yolo-system:state":{"hostname":"foo"}}')

    def test_data_from_dict_rpc(self):
        dnode = self.ctx.create_data_path("/yolo-system:format-disk")
        self.assertIsInstance(dnode, DRpc)
        dnode.merge_data_dict(
            {"duration": 42},
            strict=True,
            validate=True,
            rpcreply=True,
        )
        try:
            j = dnode.print_mem("json", pretty=False)
        finally:
            dnode.free()
        self.assertEqual(j, '{"yolo-system:format-disk":{"duration":42}}')

    def test_data_from_dict_action(self):
        module = self.ctx.get_module("yolo-system")
        dnode = module.parse_data_dict(
            {
                "conf": {
                    "url": [
                        {
                            "proto": "https",
                            "host": "github.com",
                            "fetch": {"timeout": 42},
                        },
                    ],
                },
            },
            strict=True,
            rpc=True,
        )
        self.assertIsInstance(dnode, DContainer)
        try:
            j = dnode.print_mem("json")
        finally:
            dnode.free()
        self.assertEqual(
            j,
            """{
  "yolo-system:conf": {
    "url": [
      {
        "proto": "https",
        "host": "github.com",
        "fetch": {
          "timeout": 42
        }
      }
    ]
  }
}
""",
        )

    def test_data_to_dict_action(self):
        module = self.ctx.get_module("yolo-system")
        request = module.parse_data_dict(
            {
                "conf": {
                    "url": [
                        {
                            "proto": "https",
                            "host": "github.com",
                            "fetch": {"timeout": 42},
                        },
                    ],
                },
            },
            strict=True,
            rpc=True,
        )
        request = request.find_path(
            "/yolo-system:conf/url[proto='https'][host='github.com']/fetch"
        )
        dnode = self.ctx.parse_op_mem(
            "json",
            '{"yolo-system:result":"not found"}',
            dtype=DataType.REPLY_YANG,
            parent=request,
        )
        try:
            dic = dnode.print_dict()
        finally:
            request.free()

        self.assertEqual(
            dic,
            {
                "conf": {
                    "url": [
                        {
                            "proto": "https",
                            "host": "github.com",
                            "fetch": {
                                "result": "not found",
                                "timeout": 42,  # probably bug in linyang, this is part of input
                            },
                        },
                    ],
                },
            },
        )

    DICT_NOTIF = {
        "alarm-triggered": {"description": "An error occurred", "severity": 3}
    }

    JSON_NOTIF = """{
  "yolo-system:alarm-triggered": {
    "description": "An error occurred",
    "severity": 3
  }
}
"""

    def test_notification_from_dict_module(self):
        module = self.ctx.get_module("yolo-system")
        dnotif = module.parse_data_dict(self.DICT_NOTIF, strict=True, notification=True)
        self.assertIsInstance(dnotif, DNotif)
        try:
            j = dnotif.print_mem("json")
        finally:
            dnotif.free()
        self.assertEqual(json.loads(j), json.loads(self.JSON_NOTIF))

    DICT_NOTIF_KEYLESS_LIST = {
        "config-change": {"edit": [{"target": "a"}, {"target": "b"}]},
    }

    def test_data_to_dict_keyless_list(self):
        module = self.ctx.get_module("yolo-system")
        dnotif = module.parse_data_dict(
            self.DICT_NOTIF_KEYLESS_LIST, strict=True, notification=True
        )
        self.assertIsInstance(dnotif, DNotif)
        try:
            dic = dnotif.print_dict()
        finally:
            dnotif.free()
        self.assertEqual(dic, self.DICT_NOTIF_KEYLESS_LIST)

    XML_DIFF_STATE1 = """<state xmlns="urn:yang:yolo:system">
  <hostname>foo</hostname>
  <speed>1234</speed>
  <number>1000</number>
  <number>2000</number>
  <number>3000</number>
  <url>
    <proto>https</proto>
    <host>github.com</host>
    <path>/CESNET/libyang-python</path>
    <enabled>false</enabled>
  </url>
  <url>
    <proto>http</proto>
    <host>foobar.com</host>
    <port>8080</port>
    <path>/index.html</path>
    <enabled>true</enabled>
  </url>
</state>
"""
    XML_DIFF_STATE2 = """<state xmlns="urn:yang:yolo:system">
  <hostname>foo</hostname>
  <speed>5432</speed>
  <number>1000</number>
  <number>3000</number>
  <url>
    <proto>https</proto>
    <host>github.com</host>
    <path>/CESNET/libyang-python</path>
    <enabled>true</enabled>
  </url>
  <url>
    <proto>http</proto>
    <host>foobar.com</host>
    <port>8080</port>
    <path>/index.html</path>
    <enabled>false</enabled>
  </url>
  <url>
    <proto>ftp</proto>
    <host>github.com</host>
    <path>/CESNET/libyang-python</path>
    <enabled>false</enabled>
  </url>
</state>
"""

    XML_DIFF_RESULT = """<state xmlns="urn:yang:yolo:system" xmlns:yang="urn:ietf:params:xml:ns:yang:1" yang:operation="none">
  <url>
    <proto>https</proto>
    <host>github.com</host>
    <enabled yang:operation="replace" yang:orig-default="false" yang:orig-value="false">true</enabled>
  </url>
  <url yang:operation="none">
    <proto>http</proto>
    <host>foobar.com</host>
    <enabled yang:operation="replace" yang:orig-default="false" yang:orig-value="true">false</enabled>
  </url>
  <url yang:operation="create" yang:key="[proto='http'][host='foobar.com']">
    <proto>ftp</proto>
    <host>github.com</host>
    <path>/CESNET/libyang-python</path>
    <enabled>false</enabled>
  </url>
  <number yang:operation="delete" yang:orig-position="1">2000</number>
  <speed yang:operation="replace" yang:orig-default="false" yang:orig-value="1234">5432</speed>
</state>
"""

    def test_data_diff(self):
        dnode1 = self.ctx.parse_data_mem(
            self.XML_DIFF_STATE1, "xml", validate_present=True
        )
        self.assertIsInstance(dnode1, DContainer)
        dnode2 = self.ctx.parse_data_mem(
            self.XML_DIFF_STATE2, "xml", validate_present=True
        )
        self.assertIsInstance(dnode2, DContainer)

        result = dnode1.diff(dnode2)
        self.assertEqual(result.print_mem("xml"), self.XML_DIFF_RESULT)
        dnode1.free()
        dnode2.free()

    TREE = [
        "/yolo-system:conf",
        "/yolo-system:conf/hostname",
        "/yolo-system:conf/url[proto='http'][host='foobar.com']",
        "/yolo-system:conf/url[proto='http'][host='foobar.com']/proto",
        "/yolo-system:conf/url[proto='http'][host='foobar.com']/host",
        "/yolo-system:conf/url[proto='http'][host='foobar.com']/port",
        "/yolo-system:conf/url[proto='http'][host='foobar.com']/path",
        "/yolo-system:conf/url[proto='http'][host='foobar.com']/enabled",
        "/yolo-system:conf/url[proto='https'][host='github.com']",
        "/yolo-system:conf/url[proto='https'][host='github.com']/proto",
        "/yolo-system:conf/url[proto='https'][host='github.com']/host",
        "/yolo-system:conf/url[proto='https'][host='github.com']/path",
        "/yolo-system:conf/url[proto='https'][host='github.com']/enabled",
        "/yolo-system:conf/number[.='1000']",
        "/yolo-system:conf/number[.='2000']",
        "/yolo-system:conf/number[.='3000']",
        "/yolo-system:conf/speed",
    ]

    def test_iter_tree(self):
        dnode = self.ctx.parse_data_mem(self.JSON_CONFIG, "json", validate_present=True)
        try:
            paths = [d.path() for d in dnode.iter_tree()]
            self.assertEqual(paths, self.TREE)
        finally:
            dnode.free()

    def test_find_one(self):
        dnode = self.ctx.parse_data_mem(self.JSON_CONFIG, "json", validate_present=True)
        self.assertIsInstance(dnode, DContainer)
        try:
            hostname = dnode.find_one("hostname")
            self.assertIsInstance(hostname, DNode)
            self.assertEqual(hostname.name(), "hostname")
        finally:
            dnode.free()

    def test_find_all(self):
        dnode = self.ctx.parse_data_mem(self.JSON_CONFIG, "json", validate_present=True)
        self.assertIsInstance(dnode, DContainer)
        try:
            urls = dnode.find_all("url")
            urls = list(urls)
            self.assertEqual(len(urls), 2)

            expected_url = {
                "url": [
                    {
                        "proto": "https",
                        "host": "github.com",
                        "path": "/CESNET/libyang-python",
                        "enabled": False,
                    }
                ]
            }
            self.assertEqual(urls[1].print_dict(absolute=False), expected_url)
        finally:
            dnode.free()

    def test_add_defaults(self):
        JSON = '{"yolo-nodetypes:records": [{"id": "rec1"}], "yolo-nodetypes:conf": {}}'
        dnode = self.ctx.parse_data_mem(
            JSON, "json", validate_present=True, parse_only=True
        )
        self.assertIsInstance(dnode, DList)
        node = dnode.find_one("id")
        self.assertIsInstance(node, DLeaf)
        node = dnode.find_one("name")
        self.assertIsNone(node)
        node = dnode.find_one("/yolo-system:conf/speed")
        self.assertIsNone(node)

        dnode.add_defaults(only_node=True)
        node = dnode.find_one("name")
        self.assertIsInstance(node, DLeaf)
        self.assertEqual(node.value(), "ASD")
        node = dnode.find_one("/yolo-nodetypes:conf/percentage")
        self.assertIsNone(node)
        node = dnode.find_one("/yolo-system:conf/speed")
        self.assertIsNone(node)

        dnode.add_defaults(only_module=dnode.module())
        node = dnode.find_one("/yolo-nodetypes:conf/percentage")
        self.assertIsInstance(node, DLeaf)
        self.assertEqual(node.value(), 10.2)
        node = dnode.find_one("/yolo-system:conf/speed")
        self.assertIsNone(node)

        dnode.add_defaults(only_node=False)
        node = dnode.find_path("/yolo-system:conf/speed")
        self.assertIsInstance(node, DLeaf)
        self.assertEqual(node.value(), 4321)

    def test_dnode_double_free(self):
        dnode = self.ctx.parse_data_mem(self.JSON_CONFIG, "json", validate_present=True)
        dnode.free()
        dnode.free()

    def test_dnode_unlink(self):
        dnode = self.ctx.parse_data_mem(self.JSON_CONFIG, "json", validate_present=True)
        self.assertIsInstance(dnode, DContainer)
        try:
            child = dnode.find_one("hostname")
            self.assertIsInstance(child, DNode)
            child.unlink(with_siblings=False)
            self.assertIsNone(dnode.find_one("hostname"))
            child = next(dnode.children(), None)
            self.assertIsNot(child, None)
            child.unlink(with_siblings=True)
            child = next(dnode.children(), None)
            self.assertIsNone(child, None)
        finally:
            dnode.free()

    def test_dnode_insert_sibling(self):
        MAIN = {"yolo-nodetypes:conf": {"percentage": "20.2"}}
        SIBLING = {"yolo-nodetypes:test1": 10}
        module = self.ctx.get_module("yolo-nodetypes")
        dnode1 = dict_to_dnode(MAIN, module, None, validate=False)
        dnode2 = dict_to_dnode(SIBLING, module, None, validate=False)
        self.assertEqual(len(list(dnode1.siblings(include_self=False))), 0)
        self.assertEqual(len(list(dnode2.siblings(include_self=False))), 0)
        dnode2.insert_sibling(dnode1)
        self.assertEqual(len(list(dnode1.siblings(include_self=False))), 1)
        self.assertEqual(len(list(dnode2.siblings(include_self=False))), 1)
        sibling = next(dnode1.siblings(include_self=False), None)
        self.assertIsInstance(sibling, DLeaf)
        self.assertEqual(sibling.cdata, dnode2.cdata)

    def test_dnode_insert_sibling_before_after(self):
        R1 = {"yolo-nodetypes:records": [{"id": "id1", "name": "name1"}]}
        R2 = {"yolo-nodetypes:records": [{"id": "id2", "name": "name2"}]}
        R3 = {"yolo-nodetypes:records": [{"id": "id3", "name": "name3"}]}
        module = self.ctx.get_module("yolo-nodetypes")
        dnode1 = dict_to_dnode(R1, module, None, validate=False)
        dnode2 = dict_to_dnode(R2, module, None, validate=False)
        dnode3 = dict_to_dnode(R3, module, None, validate=False)
        self.assertEqual(dnode1.first_sibling().cdata, dnode1.cdata)
        dnode1.insert_before(dnode2)
        dnode1.insert_after(dnode3)
        self.assertEqual(
            [dnode2.cdata, dnode1.cdata, dnode3.cdata],
            [s.cdata for s in dnode1.first_sibling().siblings()],
        )
        self.assertEqual(dnode1.first_sibling().cdata, dnode2.cdata)

    def _create_opaq_hostname(self):
        root = self.ctx.create_data_path(path="/yolo-system:conf")
        root.new_path(
            "hostname",
            None,
            opt_opaq=True,
        )
        return root.find_one("/yolo-system:conf/hostname")

    def test_dnode_new_opaq_find_one(self):
        dnode = self._create_opaq_hostname()

        self.assertIsInstance(dnode, DLeaf)

    def test_dnode_attrs(self):
        dnode = self._create_opaq_hostname()
        attrs = dnode.attrs()

        self.assertIsInstance(attrs, DNodeAttrs)

    def test_dnode_attrs_set(self):
        dnode = self._create_opaq_hostname()
        attrs = dnode.attrs()

        self.assertEqual(len(attrs.cdata), 0)
        attrs.set("ietf-netconf:operation", "remove")

        self.assertEqual(len(attrs.cdata), 1)

    def test_dnode_attrs_get(self):
        dnode = self._create_opaq_hostname()
        attrs = dnode.attrs()

        attrs.set("ietf-netconf:operation", "remove")

        value = attrs.get("ietf-netconf:operation")
        self.assertEqual(value, "remove")

    def test_dnode_attrs__len(self):
        dnode = self._create_opaq_hostname()
        attrs = dnode.attrs()

        self.assertEqual(len(attrs), 0)
        attrs.set("ietf-netconf:operation", "remove")

        self.assertEqual(len(attrs), 1)

    def test_dnode_attrs__contains(self):
        dnode = self._create_opaq_hostname()
        attrs = dnode.attrs()

        attrs.set("ietf-netconf:operation", "remove")

        self.assertTrue("ietf-netconf:operation" in attrs)

    def test_dnode_attrs_remove(self):
        dnode = self._create_opaq_hostname()
        attrs = dnode.attrs()

        attrs.set("ietf-netconf:operation", "remove")
        attrs.remove("ietf-netconf:operation")

        self.assertEqual(len(attrs), 0)

    def test_dnode_attrs_set_and_remove_multiple(self):
        dnode = self._create_opaq_hostname()
        attrs = dnode.attrs()

        attrs.set("ietf-netconf:operation", "remove")
        attrs.set("something:else", "test")
        attrs.set("no_prefix", "test")
        self.assertEqual(len(attrs), 3)

        attrs.remove("something:else")
        self.assertEqual(len(attrs), 2)
        self.assertIn("no_prefix", attrs)
        self.assertIn("ietf-netconf:operation", attrs)

        attrs.remove("no_prefix")
        self.assertEqual(len(attrs), 1)

        attrs.remove("ietf-netconf:operation")
        self.assertEqual(len(attrs), 0)

    def test_dnode_leafref_linking(self):
        MAIN = """{
            "yolo-leafref-extended:list1": [{
                "leaf1": "val1",
                "leaflist2": ["val2", "val3"]
            }],
            "yolo-leafref-extended:ref1": "val1"
            }"""
        self.ctx.destroy()
        self.ctx = Context(
            YANG_DIR, leafref_extended=True, leafref_linking=True, compile_obsolete=True
        )
        mod = self.ctx.load_module("yolo-leafref-extended")
        self.assertIsInstance(mod, Module)
        dnode1 = self.ctx.parse_data_mem(MAIN, "json", parse_only=True)
        self.assertIsInstance(dnode1, DList)
        dnode2 = next(dnode1.siblings(include_self=False))
        self.assertIsInstance(dnode2, DLeaf)
        dnode3 = next(dnode1.children())
        self.assertIsInstance(dnode3, DLeaf)
        self.assertIsNone(next(dnode3.leafref_nodes(), None))
        dnode2.leafref_link_node_tree()
        dnode4 = next(dnode3.leafref_nodes())
        self.assertIsInstance(dnode4, DLeaf)
        self.assertEqual(dnode4.cdata, dnode2.cdata)
        dnode5 = next(dnode4.target_nodes())
        self.assertIsInstance(dnode5, DLeaf)
        self.assertEqual(dnode5.cdata, dnode3.cdata)
        dnode1.free()

    def test_dnode_store_only(self):
        MAIN = {"yolo-nodetypes:test1": 50}
        module = self.ctx.load_module("yolo-nodetypes")
        dnode = dict_to_dnode(MAIN, module, None, validate=False, store_only=True)
        self.assertIsInstance(dnode, DLeaf)
        self.assertEqual(dnode.value(), 50)
        dnode.free()

    def test_dnode_builtin_plugins_only(self):
        MAIN = {"yolo-nodetypes:ip-address": "test"}
        self.tearDown()
        gc.collect()
        self.ctx = Context(YANG_DIR, builtin_plugins_only=True, compile_obsolete=True)
        module = self.ctx.load_module("yolo-nodetypes")
        dnode = dict_to_dnode(MAIN, module, None, validate=False, store_only=True)
        self.assertIsInstance(dnode, DLeaf)
        self.assertEqual(dnode.value(), "test")
        dnode.free()

    def test_merge_store_only(self):
        MAIN = {"yolo-nodetypes:test1": 50}
        module = self.ctx.load_module("yolo-nodetypes")
        dnode = module.parse_data_dict(MAIN, validate=False, store_only=True)
        self.assertIsInstance(dnode, DLeaf)
        self.assertEqual(dnode.value(), 50)
        dnode.free()

    def test_merge_builtin_plugins_only(self):
        MAIN = {"yolo-nodetypes:ip-address": "test"}
        self.tearDown()
        gc.collect()
        self.ctx = Context(YANG_DIR, builtin_plugins_only=True, compile_obsolete=True)
        module = self.ctx.load_module("yolo-nodetypes")
        dnode = module.parse_data_dict(MAIN, validate=False, store_only=True)
        self.assertIsInstance(dnode, DLeaf)
        self.assertEqual(dnode.value(), "test")
        dnode.free()

    def test_dnode_parse_json_null(self):
        JSON = """{"yolo-nodetypes:ip-address": null}"""
        dnode = self.ctx.parse_data_mem(JSON, "json", json_null=True)
        dnode_names = [d.name() for d in dnode.siblings()]
        self.assertFalse("ip-address" in dnode_names)



# Copyright (c) 2018-2019 Robin Jarry
# Copyright (c) 2020 6WIND S.A.
# Copyright (c) 2021 RACOM s.r.o.
# SPDX-License-Identifier: MIT

from typing import Callable, Optional

from _libyang import ffi, lib
from .context import Context
from .log import get_libyang_level
from .schema import ExtensionCompiled, ExtensionParsed, Module, SNode
from .util import LibyangError, c2str, str2c


# -------------------------------------------------------------------------------------
extensions_plugins = {}


class LibyangExtensionError(LibyangError):
    def __init__(self, message: str, ret: int, log_level: int) -> None:
        super().__init__(message)
        self.ret = ret
        self.log_level = log_level


@ffi.def_extern(name="lypy_lyplg_ext_parse_clb")
def libyang_c_lyplg_ext_parse_clb(pctx, pext):
    plugin = extensions_plugins[lib.lysc_get_ext_plugin(pext.plugin_ref)]
    module_cdata = lib.lyplg_ext_parse_get_cur_pmod(pctx).mod
    context = Context(cdata=module_cdata.ctx)
    module = Module(context, module_cdata)
    parsed_ext = ExtensionParsed(context, pext, module)
    plugin.set_parse_ctx(pctx)
    try:
        plugin.parse_clb(module, parsed_ext)
        return lib.LY_SUCCESS
    except LibyangExtensionError as e:
        ly_level = get_libyang_level(e.log_level)
        if ly_level is None:
            ly_level = lib.LY_EPLUGIN
        e_str = str(e)
        plugin.add_error_message(e_str)
        lib.lyplg_ext_parse_log(pctx, pext, ly_level, e.ret, str2c(e_str))
        return e.ret


@ffi.def_extern(name="lypy_lyplg_ext_compile_clb")
def libyang_c_lyplg_ext_compile_clb(cctx, pext, cext):
    plugin = extensions_plugins[lib.lysc_get_ext_plugin(pext.plugin_ref)]
    context = Context(cdata=lib.lyplg_ext_compile_get_ctx(cctx))
    module = Module(context, cext.module)
    parsed_ext = ExtensionParsed(context, pext, module)
    compiled_ext = ExtensionCompiled(context, cext)
    plugin.set_compile_ctx(cctx)
    try:
        plugin.compile_clb(parsed_ext, compiled_ext)
        return lib.LY_SUCCESS
    except LibyangExtensionError as e:
        ly_level = get_libyang_level(e.log_level)
        if ly_level is None:
            ly_level = lib.LY_EPLUGIN
        e_str = str(e)
        plugin.add_error_message(e_str)
        lib.lyplg_ext_compile_log(cctx, cext, ly_level, e.ret, str2c(e_str))
        return e.ret


@ffi.def_extern(name="lypy_lyplg_ext_parse_free_clb")
def libyang_c_lyplg_ext_parse_free_clb(ctx, pext):
    plugin = extensions_plugins[lib.lysc_get_ext_plugin(pext.plugin_ref)]
    context = Context(cdata=ctx)
    parsed_ext = ExtensionParsed(context, pext, None)
    plugin.parse_free_clb(parsed_ext)


@ffi.def_extern(name="lypy_lyplg_ext_compile_free_clb")
def libyang_c_lyplg_ext_compile_free_clb(ctx, cext):
    plugin = extensions_plugins[
        lib.lysc_get_ext_plugin(getattr(cext, "def").plugin_ref)
    ]
    context = Context(cdata=ctx)
    compiled_ext = ExtensionCompiled(context, cext)
    plugin.compile_free_clb(compiled_ext)


class ExtensionPlugin:
    ERROR_SUCCESS = lib.LY_SUCCESS
    ERROR_MEM = lib.LY_EMEM
    ERROR_INVALID_INPUT = lib.LY_EINVAL
    ERROR_NOT_VALID = lib.LY_EVALID
    ERROR_DENIED = lib.LY_EDENIED
    ERROR_NOT = lib.LY_ENOT

    def __init__(
        self,
        module_name: str,
        name: str,
        id_str: str,
        context: Optional[Context] = None,
        parse_clb: Optional[Callable[[Module, ExtensionParsed], None]] = None,
        compile_clb: Optional[
            Callable[[ExtensionParsed, ExtensionCompiled], None]
        ] = None,
        parse_free_clb: Optional[Callable[[ExtensionParsed], None]] = None,
        compile_free_clb: Optional[Callable[[ExtensionCompiled], None]] = None,
    ) -> None:
        """
        Set the callback functions, which will be called if libyang will be processing
        given extension defined by name from module defined by module_name.

        :arg self:
            This instance of extension plugin
        :arg module_name:
            The name of module in which the extension is defined
        :arg name:
            The name of extension itself
        :arg id_str:
            The unique ID of extension plugin within the libyang context
        :arg context:
            The context in which the extension plugin will be used. If set to None,
            the plugin will be used for all existing and even future contexts
        :arg parse_clb:
            The optional callback function of which will be called during extension parsing
            Expected arguments are:
                module: The module which is being parsed
                extension: The exact extension instance
            Expected raises:
                LibyangExtensionError in case of processing error
        :arg compile_clb:
            The optional callback function of which will be called during extension compiling
            Expected arguments are:
                extension_parsed: The parsed extension instance
                extension_compiled: The compiled extension instance
            Expected raises:
                LibyangExtensionError in case of processing error
        :arg parse_free_clb
            The optional callback function of which will be called during freeing of parsed extension
            Expected arguments are:
                extension: The parsed extension instance to be freed
        :arg compile_free_clb
            The optional callback function of which will be called during freeing of compiled extension
            Expected arguments are:
                extension: The compiled extension instance to be freed
        """
        self.context = context
        self.module_name = module_name
        self.module_name_cstr = str2c(self.module_name)
        self.name = name
        self.name_cstr = str2c(self.name)
        self.id_str = id_str
        self.id_cstr = str2c(self.id_str)
        self.parse_clb = parse_clb
        self.compile_clb = compile_clb
        self.parse_free_clb = parse_free_clb
        self.compile_free_clb = compile_free_clb
        self._error_messages = []
        self._pctx = ffi.NULL
        self._cctx = ffi.NULL

        self.cdata = ffi.new("struct lyplg_ext_record[2]")
        self.cdata[0].module = self.module_name_cstr
        self.cdata[0].name = self.name_cstr
        self.cdata[0].plugin.id = self.id_cstr
        if self.parse_clb is not None:
            self.cdata[0].plugin.parse = lib.lypy_lyplg_ext_parse_clb
        if self.compile_clb is not None:
            self.cdata[0].plugin.compile = lib.lypy_lyplg_ext_compile_clb
        if self.parse_free_clb is not None:
            self.cdata[0].plugin.pfree = lib.lypy_lyplg_ext_parse_free_clb
        if self.compile_free_clb is not None:
            self.cdata[0].plugin.cfree = lib.lypy_lyplg_ext_compile_free_clb
        ret = lib.lyplg_add_extension_plugin(
            context.cdata if context is not None else ffi.NULL,
            lib.LYPLG_EXT_API_VERSION,
            ffi.cast("const void *", self.cdata),
        )
        if ret != lib.LY_SUCCESS:
            raise LibyangError("Unable to add extension plugin")
        if self.cdata[0].plugin not in extensions_plugins:
            extensions_plugins[self.cdata[0].plugin] = self

    def __del__(self) -> None:
        if self.cdata[0].plugin in extensions_plugins:
            del extensions_plugins[self.cdata[0].plugin]

    @staticmethod
    def stmt2str(stmt: int) -> str:
        return c2str(lib.lyplg_ext_stmt2str(stmt))

    def add_error_message(self, err_msg: str) -> None:
        self._error_messages.append(err_msg)

    def clear_error_messages(self) -> None:
        self._error_messages.clear()

    def set_parse_ctx(self, pctx) -> None:
        self._pctx = pctx

    def set_compile_ctx(self, cctx) -> None:
        self._cctx = cctx

    def parse_substmts(self, ext: ExtensionParsed) -> int:
        return lib.lyplg_ext_parse_extension_instance(self._pctx, ext.cdata)

    def compile_substmts(
        self,
        pext: ExtensionParsed,
        cext: ExtensionCompiled,
        parent: Optional[SNode] = None,
    ) -> int:
        return lib.lyplg_ext_compile_extension_instance(
            self._cctx,
            pext.cdata,
            cext.cdata,
            ffi.NULL if parent is None else parent.cdata,
        )

    def free_parse_substmts(self, ext: ExtensionParsed) -> None:
        lib.lyplg_ext_pfree_instance_substatements(
            self.context.cdata, ext.cdata.substmts
        )

    def free_compile_substmts(self, ext: ExtensionCompiled) -> None:
        lib.lyplg_ext_cfree_instance_substatements(
            self.context.cdata, ext.cdata.substmts
        )



# Copyright (c) 2018-2019 Robin Jarry
# SPDX-License-Identifier: MIT

import os
import unittest

from libyang import Context, LibyangError, Module, SContainer, SLeaf, SLeafList
from libyang.util import c2str


YANG_DIR = os.path.join(os.path.dirname(__file__), "yang")


# -------------------------------------------------------------------------------------


class ContextTest(unittest.TestCase):
    def test_ctx_no_dir(self):
        with Context() as ctx:
            self.assertIsNot(ctx, None)

    def test_ctx_yanglib(self):
        ctx = Context(
            YANG_DIR,
            yanglib_path=YANG_DIR + "/yang-library.json",
            compile_obsolete=True,
        )
        ctx.load_module("yolo-system")
        dnode = ctx.get_yanglib_data()
        j = dnode.print_mem("json", with_siblings=True)
        self.assertIsInstance(j, str)

    def test_ctx_dir(self):
        with Context(YANG_DIR) as ctx:
            self.assertIsNot(ctx, None)

    def test_ctx_duplicate_searchpath(self):
        duplicate_search_path = ":".join([YANG_DIR, YANG_DIR])
        try:
            Context(duplicate_search_path)
        except LibyangError:
            self.fail("Context.__init__ should not raise LibyangError")

    def test_ctx_invalid_dir(self):
        with Context("/does/not/exist") as ctx:
            self.assertIsNot(ctx, None)

    def test_ctx_missing_dir(self):
        with Context(os.path.join(YANG_DIR, "yolo")) as ctx:
            self.assertIsNot(ctx, None)
            with self.assertRaises(LibyangError):
                ctx.load_module("yolo-system")

    def test_ctx_env_search_dir(self):
        try:
            os.environ["YANGPATH"] = ":".join(
                [os.path.join(YANG_DIR, "omg"), os.path.join(YANG_DIR, "wtf")]
            )
            with Context(os.path.join(YANG_DIR, "yolo")) as ctx:
                mod = ctx.load_module("yolo-system")
                self.assertIsInstance(mod, Module)
        finally:
            del os.environ["YANGPATH"]

    def test_ctx_load_module(self):
        with Context(YANG_DIR) as ctx:
            mod = ctx.load_module("yolo-system")
            self.assertIsInstance(mod, Module)

    def test_ctx_load_module_with_features(self):
        with Context(YANG_DIR) as ctx:
            mod = ctx.load_module("yolo-system", None, ["*"])
            self.assertIsInstance(mod, Module)
            for f in list(mod.features()):
                self.assertTrue(f.state())

    def test_ctx_get_module(self):
        with Context(YANG_DIR) as ctx:
            ctx.load_module("yolo-system")
            mod = ctx.get_module("wtf-types")
            self.assertIsInstance(mod, Module)

    def test_ctx_get_invalid_module(self):
        with Context(YANG_DIR) as ctx:
            ctx.load_module("wtf-types")
            with self.assertRaises(LibyangError):
                ctx.get_module("yolo-system")

    def test_ctx_load_invalid_module(self):
        with Context(YANG_DIR) as ctx:
            with self.assertRaises(LibyangError):
                ctx.load_module("invalid-module")

    def test_ctx_find_path(self):
        with Context(YANG_DIR) as ctx:
            ctx.load_module("yolo-system")
            node = next(ctx.find_path("/yolo-system:conf/offline"))
            self.assertIsInstance(node, SLeaf)
            node2 = next(ctx.find_path("../number", root_node=node))
            self.assertIsInstance(node2, SLeafList)

    def test_ctx_find_xpath_atoms(self):
        with Context(YANG_DIR) as ctx:
            ctx.load_module("yolo-system")
            node_iter = ctx.find_xpath_atoms("/yolo-system:conf/offline")
            node = next(node_iter)
            self.assertIsInstance(node, SContainer)
            node = next(node_iter)
            self.assertIsInstance(node, SLeaf)
            node_iter = ctx.find_xpath_atoms("../number", root_node=node)
            node = next(node_iter)
            self.assertIsInstance(node, SLeaf)
            node = next(node_iter)
            self.assertIsInstance(node, SContainer)
            node = next(node_iter)
            self.assertIsInstance(node, SLeafList)

    def test_ctx_iter_modules(self):
        with Context(YANG_DIR) as ctx:
            ctx.load_module("yolo-system")
            modules = list(iter(ctx))
            self.assertGreater(len(modules), 0)

    YOLO_MOD_PATH = os.path.join(YANG_DIR, "yolo/yolo-system.yang")

    def test_ctx_parse_module(self):
        with open(self.YOLO_MOD_PATH, encoding="utf-8") as f:
            mod_str = f.read()
        with Context(YANG_DIR) as ctx:
            mod = ctx.parse_module_str(mod_str, features=["turbo-boost", "networking"])
            self.assertIsInstance(mod, Module)

        with open(self.YOLO_MOD_PATH, encoding="utf-8") as f:
            with Context(YANG_DIR) as ctx:
                mod = ctx.parse_module_file(f, features=["turbo-boost", "networking"])
                self.assertIsInstance(mod, Module)

    def test_ctx_leafref_extended(self):
        with Context(YANG_DIR, leafref_extended=True) as ctx:
            mod = ctx.load_module("yolo-leafref-extended")
            self.assertIsInstance(mod, Module)

    def test_context_dict(self):
        with Context(YANG_DIR) as ctx:
            orig_str = "teststring"
            handle = ctx.add_to_dict(orig_str)
            self.assertEqual(orig_str, c2str(handle))
            ctx.remove_from_dict(orig_str)

    def test_ctx_disable_searchdirs(self):
        with Context(YANG_DIR, disable_searchdirs=True) as ctx:
            with self.assertRaises(LibyangError):
                ctx.load_module("yolo-nodetypes")

    def test_ctx_using_clb(self):
        def get_module_valid_clb(mod_name, *_):
            YOLO_NODETYPES_MOD_PATH = os.path.join(YANG_DIR, "yolo/yolo-nodetypes.yang")
            self.assertEqual(mod_name, "yolo-nodetypes")
            with open(YOLO_NODETYPES_MOD_PATH, encoding="utf-8") as f:
                mod_str = f.read()
            return "yang", mod_str

        def get_module_invalid_clb(mod_name, *_):
            return None

        with Context(YANG_DIR, disable_searchdirs=True) as ctx:
            with self.assertRaises(LibyangError):
                ctx.load_module("yolo-nodetypes")

            ctx.external_module_loader.set_module_data_clb(get_module_invalid_clb)
            with self.assertRaises(LibyangError):
                mod = ctx.load_module("yolo-nodetypes")

            ctx.external_module_loader.set_module_data_clb(get_module_valid_clb)
            mod = ctx.load_module("yolo-nodetypes")
            self.assertIsInstance(mod, Module)



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: MIT

import contextlib
import fnmatch
import re
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from .keyed_list import KeyedList, py_to_yang


# -------------------------------------------------------------------------------------
XPATH_ELEMENT_RE = re.compile(r"^/(?:(?P<prefix>[-\w]+):)?(?P<name>[-\w\*]+)")


def xpath_split(xpath: str) -> Iterator[Tuple[str, str, List[Tuple[str, str]]]]:
    """
    Return an iterator that yields xpath components:

        (prefix, name, keys)

    Where:

        :var prefix:
            The YANG prefix where ``name`` is defined. May be ``None`` if no prefix is
            specified.
        :var name:
            The name of the YANG element (container, list, leaf, leaf-list).
        :var keys:
            A list of tuples ``("key_name", "key_value")`` parsed from Xpath key
            specifiers: ``[key_name="key_value"]...``.

    Example:

    >>> list(xpath_split("/p1:a/b/c[k='v']/p2:d"))
    [("p1", "a", []), (None, "b", []), (None, "c", [("k", "v")]), ("p2", "d", [])]
    """
    xpath = xpath.strip()
    if not xpath:
        raise ValueError("empty xpath")
    if xpath[0] != "/":
        # support relative xpaths
        xpath = "/" + xpath
    i = 0

    while i < len(xpath):
        match = XPATH_ELEMENT_RE.search(xpath[i:])
        if not match:
            raise ValueError(
                "invalid xpath: %r (expected r'%s' at index %d)"
                % (xpath, XPATH_ELEMENT_RE.pattern, i)
            )
        prefix, name = match.groups()
        i += match.end()

        keys = []
        while i < len(xpath) and xpath[i] == "[":
            i += 1  # skip opening '['
            j = xpath.find("=", i)  # find key name end

            if j != -1:  # keyed specifier
                key_name = xpath[i:j]
                quote = xpath[j + 1]  # record opening quote character
                j = i = j + 2  # skip '=' and opening quote
                while True:
                    if xpath[j] == quote and xpath[j - 1] != "\\":
                        break
                    j += 1
                # replace escaped chars by their non-escape version
                key_value = xpath[i:j].replace(f"\\{quote}", f"{quote}")
                keys.append((key_name, key_value))
                i = j + 2  # skip closing quote and ']'
            else:  # index specifier
                j = i
                while True:
                    if xpath[j] == "]":
                        break
                    j += 1
                key_value = xpath[i:j]
                keys.append(("", key_value))
                i = j + 2

        yield prefix, name, keys


# -------------------------------------------------------------------------------------
def _xpath_keys_to_key_name(
    keys: List[Tuple[str, str]],
) -> Optional[Union[str, Tuple[str, ...]]]:
    """
    Extract key name from parsed xpath keys returned by xpath_split. The return value
    of this function can be used as key_name argument when creating a new KeyedList
    object.

    Examples:

    >>> _xpath_keys_to_key_name([('.', 'foo')])
    # -> None
    >>> _xpath_keys_to_key_name([('name', 'baz')])
    'name'
    >>> _xpath_keys_to_key_name([('first-name', 'Charly'), ('last-name', 'Oleg')])
    ('first-name', 'last-name')
    """
    if keys[0][0] == ".":
        # leaf-list
        return None
    if len(keys) == 1:
        return keys[0][0]
    return tuple(k[0] for k in keys)


# -------------------------------------------------------------------------------------
def _xpath_keys_to_key_val(keys: List[Tuple[str, str]]) -> Union[str, Tuple[str, ...]]:
    """
    Extract key value from parsed xpath keys returned by xpath_split. The return value
    of this function can be used as argument to lookup elements in a KeyedList object.

    Examples:

    >>> _xpath_keys_to_key_val([('.', 'foo')])
    'foo'
    >>> _xpath_keys_to_key_val([('name', 'baz')])
    'baz'
    >>> _xpath_keys_to_key_val([('first-name', 'Charly'), ('last-name', 'Oleg')])
    ('Charly', 'Oleg')
    """
    if len(keys) == 1:
        return keys[0][1]
    return tuple(k[1] for k in keys)


# -------------------------------------------------------------------------------------
def _list_find_key_index(keys: List[Tuple[str, str]], lst: List) -> int:
    """
    Find the index of an element matching the parsed xpath keys returned by xpath_split
    into a native python list.

    :raises ValueError:
        If the element is not found.
    :raises TypeError:
        If the list elements are not dictionaries.
    """
    if keys[0][0] == ".":
        # leaf-list
        for i, elem in enumerate(lst):
            if py_to_yang(elem) == keys[0][1]:
                return i

    elif keys[0][0] == "":
        # keys[0][1] is directly the index
        index = int(keys[0][1]) - 1
        if len(lst) > index:
            return index

    else:
        for i, elem in enumerate(lst):
            if not isinstance(elem, dict):
                raise TypeError("expected a dict")
            if all(k in elem and py_to_yang(elem[k]) == v for k, v in keys):
                return i

    raise ValueError("%s not found in list" % keys)


# -------------------------------------------------------------------------------------
def _xpath_find(data: Dict, xparts: List, create_if_missing: bool = False) -> Any:
    """
    Descend into a data dictionary.

    :arg data:
        The dictionary where to look for `xparts`.
    :arg xparts:
        Elements of an Xpath split with xpath_split()
    :arg bool create_if_missing:
        If elements are missing from `data`, create them.

    :returns:
        The element identified by `xparts`.
    :raises KeyError:
        If create_if_missing=False and the element is not found in `data`.
    :raises TypeError:
        If `data` does not match the expected structure conveyed by `xparts`.
    """
    for _, name, keys in xparts:
        if not isinstance(data, dict):
            raise TypeError("expected a dict")
        if keys:
            if name not in data and create_if_missing:
                data[name] = KeyedList(key_name=_xpath_keys_to_key_name(keys))
            lst = data[name]  # may raise KeyError
            if isinstance(lst, KeyedList):
                try:
                    data = lst[_xpath_keys_to_key_val(keys)]
                except KeyError:
                    if not create_if_missing:
                        raise
                    data = dict(keys)
                    lst.append(data)

            elif isinstance(lst, list):
                # regular python list, need to iterate over it
                try:
                    i = _list_find_key_index(keys, lst)
                    data = lst[i]
                except ValueError:
                    # not found
                    if not create_if_missing:
                        raise KeyError(keys) from None
                    data = dict(keys)
                    lst.append(data)

            else:
                raise TypeError("expected a list")

        elif create_if_missing:
            data = data.setdefault(name, {})

        else:
            data = data[name]  # may raise KeyError

    return data


# -------------------------------------------------------------------------------------
def xpath_get(data: Dict, xpath: str, default: Any = None) -> Any:
    """
    Get an element from a data structure (dict) that matches the given xpath.

    Examples:

    >>> config = {'conf': {'net': [{'name': 'mgmt', 'routing': {'a': 1}}]}}
    >>> xpath_get(config, '/prefix:conf/net[name="mgmt"]/routing')
    {'a': 1}
    >>> xpath_get(config, '/prefix:conf/net[name="prod"]/routing')
    >>> xpath_get(config, '/prefix:conf/net[name="prod"]/routing', {})
    {}
    """
    try:
        return _xpath_find(data, xpath_split(xpath), create_if_missing=False)
    except KeyError:
        return default


# -------------------------------------------------------------------------------------
def xpath_getall(data: Dict, xpath: str) -> Iterator[Any]:
    """
    Yield all elements from a data structure (dict) that match the given
    xpath. Basic wildcards in the xpath are supported.

    IMPORTANT: the order in which the elements are yielded is not stable and
    you should not rely on it.

    Examples:

    >>> config = {'config': {'vrf': [{'name': 'vr0', 'routing': {'a1': [1, 8]}},
    ...                              {'name': 'vrf1', 'routing': {'a2': 55}},
    ...                              {'name': 'vrf2', 'snmp': {'a3': 12, 'c': 5}}]}}
    >>> list(xpath_getall(config, '/config/vrf/name'))
    ['vrf0', 'vrf1']
    >>> list(xpath_getall(config, '/config/vrf/routing'))
    [{'a1': [1, 8]}, {'a2': 55}]
    >>> list(xpath_getall(config, '/config/vrf/routing/b'))
    []
    >>> list(xpath_getall(config, '/config/vrf/*/a*'))
    [1, 8, 55, 12]
    """
    parts = list(xpath_split(xpath))

    def _walk_subtrees(subtrees, keys, level):
        next_xpath = "/" + "/".join(
            n + "".join('[%s="%s"]' % _k for _k in k) for _, n, k in parts[level:]
        )
        # pylint: disable=too-many-nested-blocks
        for sub in subtrees:
            if isinstance(sub, list):
                if keys:
                    try:
                        if isinstance(sub, KeyedList):
                            l = sub[_xpath_keys_to_key_val(keys)]
                        else:
                            l = sub[_list_find_key_index(keys, sub)]
                    except (ValueError, KeyError):
                        continue
                    if level == len(parts):
                        yield l
                    elif isinstance(l, dict):
                        yield from xpath_getall(l, next_xpath)
                else:
                    if level == len(parts):
                        yield from sub
                    else:
                        for l in sub:
                            if isinstance(l, dict):
                                yield from xpath_getall(l, next_xpath)

            elif isinstance(sub, dict) and not keys and level < len(parts):
                yield from xpath_getall(sub, next_xpath)

            elif level == len(parts) and not keys:
                yield sub

    for i, (_, name, keys) in enumerate(parts):
        if not isinstance(data, dict) or name not in data:
            if "*" in name and isinstance(data, dict):
                # Wildcard xpath element, yield from all matching subtrees
                subtrees = (data[n] for n in fnmatch.filter(data.keys(), name))
                yield from _walk_subtrees(subtrees, keys, i + 1)
            return

        data = data[name]

        if keys:
            if isinstance(data, KeyedList):
                # shortcut access is possible
                try:
                    data = data[_xpath_keys_to_key_val(keys)]
                except (KeyError, ValueError):
                    return
            elif isinstance(data, list):
                # regular python list, need to iterate over it
                try:
                    i = _list_find_key_index(keys, data)
                    data = data[i]
                except ValueError:
                    return
            else:
                return
        elif isinstance(data, list):
            # More than one element matches the xpath.
            yield from _walk_subtrees(data, keys, i + 1)
            return

    # trivial case, only one match
    yield data


# -------------------------------------------------------------------------------------
def xpath_set(
    data: Dict,
    xpath: str,
    value: Any,
    *,
    force: bool = True,
    after: Optional[str] = None,
) -> Any:
    """
    Set the value pointed by the provided xpath into the provided data structure. If
    force=False and the value is already set in the data structure, do not overwrite it
    and return the current value.

    :arg data:
        The dictionary to update.
    :arg xpath:
        The path where to insert the value.
    :arg value:
        The value to insert in data.
    :arg force:
        Overwrite if a value already exists at the given xpath.
    :arg after:
        The element after which to insert the provided value. The meaning of this
        argument is similar to the prev_list argument of sr_get_change_tree_next:

        https://github.com/sysrepo/sysrepo/blob/v1.4.66/src/sysrepo.h#L1399-L1405

        `after=None`
            Append at the end of the list if not already present.
        `after=""`
            Insert at the beginning of the list if not already present.
        `after="foo"`
            Insert after the YANG leaf-list "foo" element. If "foo" element is not
            found, raise a ValueError.
        `after="[key1='value1']..."`
            Insert after the YANG list element with matching keys. If such element is
            not found, raise a ValueError.

        If the value identified by `xpath` already exists in data, the `after` argument
        is ignored. If force=True, the existing value is replaced (at its current
        position in the list), otherwise the existing value is returned.

    :returns:
        The inserted value or the existing value if force=False and a value already
        exists a the given xpath.
    :raises ValueError:
        If `after` is specified and no matching element to insert after is found.

    Examples:

    >>> state = {'state': {'vrf': [{'name': 'vr0', 'routing': {'a': 1}}]}}
    >>> xpath_set(state, '/state/vrf[name="vr0"]/routing/b', 55)
    55
    >>> state
    {'state': {'vrf': [{'name': 'vr0', 'routing': {'a': 1, 'b': 55}}]}}
    >>> xpath_set(state, '/state/vrf[name="vr0"]/routing', {'c': 8})
    {'a': 1, 'b': 55}
    >>> state
    {'state': {'vrf': [{'name': 'vr0', 'routing': {'a': 1, 'b': 55}}]}}
    """
    parts = list(xpath_split(xpath))
    parent = _xpath_find(data, parts[:-1], create_if_missing=True)
    _, name, keys = parts[-1]

    if not keys:
        # name points to a container or leaf, trivial
        if force:
            parent[name] = value
            return value
        return parent.setdefault(name, value)

    # list or leaf-list
    if name not in parent:
        if after is not None:
            parent[name] = []
        else:
            parent[name] = KeyedList(key_name=_xpath_keys_to_key_name(keys))

    lst = parent[name]

    if isinstance(lst, KeyedList):
        # shortcut access is possible
        if after is not None:
            raise ValueError("after='...' is not supported for unordered lists")
        key_val = _xpath_keys_to_key_val(keys)
        if key_val in lst:
            if force:
                del lst[key_val]
                lst.append(value)
        else:
            lst.append(value)
        return lst[key_val]

    # regular python list from now
    if not isinstance(lst, list):
        raise TypeError("expected a list")

    with contextlib.suppress(ValueError):
        i = _list_find_key_index(keys, lst)
        # found
        if force:
            lst[i] = value
        return lst[i]

    # value not found; handle insertion based on 'after'
    if after is None:
        lst.append(value)
        return value

    if after == "":
        lst.insert(0, value)
        return value

    # first try to find the value in the leaf list
    try:
        _, _, after_keys = next(
            xpath_split(f"/*{after}" if after[0] == "[" else f"/*[.={after!r}]")
        )
        insert_index = _list_find_key_index(after_keys, lst) + 1
    except ValueError:
        # handle 'after' as numeric index
        if not after.isnumeric():
            raise

        insert_index = int(after)
        if insert_index > len(lst):
            raise

    if insert_index == len(lst):
        lst.append(value)
    else:
        lst.insert(insert_index, value)

    return value


# -------------------------------------------------------------------------------------
def xpath_setdefault(data: Dict, xpath: str, default: Any) -> Any:
    """
    Shortcut for xpath_set(..., force=False).
    """
    return xpath_set(data, xpath, default, force=False)


# -------------------------------------------------------------------------------------
def xpath_del(data: Dict, xpath: str) -> bool:
    """
    Remove an element identified by an Xpath from a data structure.

    :arg data:
        The dictionary to modify.
    :arg xpath:
        The path identifying the element to remove.

    :returns:
        True if the element was removed. False if the element was not found.
    """
    parts = list(xpath_split(xpath))
    try:
        parent = _xpath_find(data, parts[:-1], create_if_missing=False)
    except KeyError:
        return False
    _, name, keys = parts[-1]
    if name not in parent:
        return False

    if keys:
        lst = parent[name]
        if isinstance(lst, KeyedList):
            # shortcut access is possible
            try:
                del lst[_xpath_keys_to_key_val(keys)]
            except (ValueError, KeyError):
                return False
        elif isinstance(lst, list):
            # regular python list, need to iterate over it
            try:
                i = _list_find_key_index(keys, lst)
                del lst[i]
            except ValueError:
                return False
        else:
            return False
        if not lst:
            # list is now empty
            del parent[name]
    else:
        del parent[name]

    return True


# -------------------------------------------------------------------------------------
def xpath_move(data: Dict, xpath: str, after: str) -> None:
    """
    Move a list element nested into a data dictionary.

    :arg data:
        The dictionary to modify.
    :arg xpath:
        The path identifying the element to move.
    :arg after:
        The element after which to move the element identified by xpath. Similar
        semantics than in xpath_set().

    :raises ValueError:
        If `xpath` does not designate a list element or if `after` is invalid.
    :raises KeyError:
        If the element identified by `xpath` or if the element identified by `after` are
        not found.
    :raises TypeError:
        If `data` does not match the expected structure conveyed by `xpath`.
    """
    parts = list(xpath_split(xpath))
    parent = _xpath_find(data, parts[:-1], create_if_missing=False)
    _, name, keys = parts[-1]
    if name not in parent:
        raise KeyError(name)
    if not keys:
        raise ValueError("xpath does not designate a list element")

    lst = parent[name]
    if isinstance(lst, KeyedList):
        raise ValueError("cannot move elements in non-ordered lists")
    if not isinstance(lst, list):
        raise ValueError("expected a list")

    try:
        i = _list_find_key_index(keys, lst)
    except ValueError:
        raise KeyError(keys) from None

    if after is None:
        lst.append(lst.pop(i))

    elif after == "":
        lst.insert(0, lst.pop(i))

    else:
        if after[0] != "[":
            after = "[.=%r]" % after
        _, _, after_keys = next(xpath_split("/*" + after))
        try:
            j = _list_find_key_index(after_keys, lst)
        except ValueError:
            raise KeyError(after) from None
        if i > j:
            j += 1
        moved = lst.pop(i)
        if j == len(lst):
            lst.append(moved)
        else:
            lst.insert(j, moved)
